from cone.app import get_root
from cone.app import testing
from cone.app.browser.ajax import ajax_continue
from cone.app.browser.ajax import ajax_form_template
from cone.app.browser.ajax import ajax_message
from cone.app.browser.ajax import ajax_status_message
from cone.app.browser.ajax import ajax_tile
from cone.app.browser.ajax import AjaxAction
from cone.app.browser.ajax import AjaxEvent
from cone.app.browser.ajax import AjaxFormContinue
from cone.app.browser.ajax import AjaxMessage
from cone.app.browser.ajax import AjaxOverlay
from cone.app.browser.ajax import AjaxPath
from cone.app.browser.ajax import livesearch
from cone.app.browser.ajax import render_ajax_form
from cone.app.browser.form import Form
from cone.app.interfaces import ILiveSearch
from cone.tile import Tile
from cone.tile import tile
from cone.tile.tests import TileTestCase
from yafowil.base import factory
from zope.component import adapter
from zope.interface import implementer
from zope.interface import Interface
import json


class TestBrowserAjax(TileTestCase):
    layer = testing.security

    def test_ajax_tile(self):
        # ``ajax_tile`` is the server side bdajax implementation for cone.
        # Using ``bdajax.action`` with cone renders tiles by action name.
        with self.layer.hook_tile_reg():
            @tile(name='testtile')
            class TestTile(Tile):
                def render(self):
                    return 'rendered test tile'

        root = get_root()
        request = self.layer.new_request()
        request.params['bdajax.action'] = 'testtile'
        request.params['bdajax.mode'] = 'replace'
        request.params['bdajax.selector'] = '.foo'

        # Fails unauthenticated, since default permission for tiles is 'view'
        self.assertEqual(ajax_tile(root, request), {})
        self.assertEqual(request.response.status, '403 Forbidden')

        # Authenticate and test again
        with self.layer.authenticated('max'):
            res = ajax_tile(root, request)
        self.assertEqual(res, {
            'continuation': False,
            'payload': 'rendered test tile',
            'mode': 'replace',
            'selector': '.foo'
        })

        # Test with error raising tile
        with self.layer.hook_tile_reg():
            @tile(name='errortile')
            class ErrorTile(Tile):
                def render(self):
                    raise Exception('Error while rendering')

        request = self.layer.new_request()
        request.params['bdajax.action'] = 'errortile'
        request.params['bdajax.mode'] = 'replace'
        request.params['bdajax.selector'] = '.foo'

        with self.layer.authenticated('max'):
            res = ajax_tile(root, request)
        self.assertEqual(res['payload'], '')
        self.assertEqual(res['mode'], 'NONE')
        self.assertEqual(res['selector'], 'NONE')
        self.assertEqual(res['continuation'][0]['flavor'], 'error')
        self.assertEqual(res['continuation'][0]['type'], 'message')
        self.assertEqual(res['continuation'][0]['selector'], None)
        expected = 'Exception: Error while rendering'
        self.assertTrue(res['continuation'][0]['payload'].find(expected) > -1)

    def test_AjaxAction(self):
        target = 'http://example.com'
        actionname = 'tilename'
        mode = 'replace'
        selector = '.someselector'
        action = AjaxAction(target, actionname, mode, selector)
        self.assertEqual(
            (action.name, action.selector, action.mode, action.target),
            (actionname, selector, mode, target)
        )

    def test_AjaxEvent(self):
        target = 'http://example.com'
        eventname = 'contextchanged'
        selector = '.contextsensitiv'
        event = AjaxEvent(target, eventname, selector)
        self.assertEqual(
            (event.name, event.selector, event.target),
            (eventname, selector, target)
        )

    def test_AjaxMessage(self):
        payload = 'Some info message'
        flavor = 'info'
        selector = 'None'
        message = AjaxMessage(payload, flavor, selector)
        self.assertEqual(
            (message.payload, message.flavor, message.selector),
            (payload, flavor, selector)
        )

    def test_AjaxOverlay(self):
        selector = '#ajax-overlay'
        action = None
        target = None
        close = False
        content_selector = '.overlay_content'
        css = None
        overlay = AjaxOverlay()
        self.assertEqual(
            (
                overlay.selector, overlay.action, overlay.target,
                overlay.close, overlay.content_selector, overlay.css
            ),
            (selector, action, target, close, content_selector, css),
        )

        selector = '#ajax-overlay'
        action = 'someaction'
        target = 'http://example.com'
        close = False
        content_selector = '.overlay_content'
        css = 'additional-css-class'
        overlay = AjaxOverlay(
            selector=selector,
            action=action,
            target=target,
            close=close,
            content_selector=content_selector,
            css=css
        )
        self.assertEqual(
            (
                overlay.selector, overlay.action, overlay.target,
                overlay.close, overlay.content_selector, overlay.css
            ),
            (selector, action, target, close, content_selector, css),
        )

    def test_AjaxPath(self):
        path = 'foo/bar'
        target = None
        action = None
        event = None
        overlay = None
        overlay_css = None
        apath = AjaxPath(path)
        self.assertEqual(
            (
                apath.path, apath.target, apath.action, apath.event,
                apath.overlay, apath.overlay_css
            ),
            (path, target, action, event, overlay, overlay_css)
        )

        path = 'foo/bar'
        target = 'http://example.com/foo/bar'
        action = 'layout:#layout:replace'
        event = 'contextchanged:#someid'
        overlay = 'acionname:#custom-overlay:.custom_overlay_content'
        overlay_css = 'additional-overlay-css-class'
        apath = AjaxPath(
            path,
            target=target,
            action=action,
            event=event,
            overlay=overlay,
            overlay_css=overlay_css
        )
        self.assertEqual(
            (
                apath.path, apath.target, apath.action, apath.event,
                apath.overlay, apath.overlay_css
            ),
            (path, target, action, event, overlay, overlay_css)
        )

    def test_ajax_continue(self):
        with self.layer.hook_tile_reg():
            @tile(name='testtile2')
            class TestTile(Tile):
                def render(self):
                    ajax_continue(
                        self.request,
                        AjaxAction('target', 'name', 'mode', 'selector')
                    )
                    return u''

        root = get_root()
        request = self.layer.new_request()
        request.params['bdajax.action'] = 'testtile2'
        request.params['bdajax.mode'] = 'replace'
        request.params['bdajax.selector'] = '.foo'

        with self.layer.authenticated('max'):
            self.assertEqual(ajax_tile(root, request), {
                'continuation': [{
                    'mode': 'mode',
                    'selector': 'selector',
                    'type': 'action',
                    'target': 'target',
                    'name': 'name'
                }],
                'payload': u'',
                'mode': 'replace',
                'selector': '.foo'
            })

    def test_ajax_message(self):
        # ``ajax_message`` is a shortcut for settings continuation message
        request = self.layer.new_request()
        ajax_message(request, 'payload')
        messages = request.environ['cone.app.continuation']
        self.assertTrue(len(messages) == 1)
        message = messages[0]
        self.assertTrue(isinstance(message, AjaxMessage))
        self.assertEqual(
            (message.payload, message.flavor, message.selector),
            ('payload', 'message', None)
        )

    def test_ajax_status_message(self):
        # ``ajax_status_message`` is a shortcut for settings continuation
        # status message
        request = self.layer.new_request()
        ajax_status_message(request, 'payload')
        messages = request.environ['cone.app.continuation']
        self.assertTrue(len(messages) == 1)
        message = messages[0]
        self.assertTrue(isinstance(message, AjaxMessage))
        self.assertEqual(
            (message.payload, message.flavor, message.selector),
            ('payload', None, '#status_message')
        )

    def test_AjaxFormContinue(self):
        # AjaxFormContinue object. This object is used by ``render_ajax_form``
        result = ''
        continuation = []
        afc = AjaxFormContinue(result, continuation)
        self.assertEqual(afc.form, '')
        self.assertEqual(afc.next, 'false')

        # If no continuation definitions, ``form`` returns result and ``next``
        # returns 'false'
        result = 'rendered form'
        afc = AjaxFormContinue(result, [])
        self.assertEqual(afc.form, 'rendered form')
        self.assertEqual(afc.next, 'false')

        # If continuation definitions and result, ``form`` returns empty
        # string, because form processing was successful. ``next`` returns a
        # JSON dump of given actions, which gets interpreted and executed on
        # client side
        action = AjaxAction(
            target='http://example.com',
            name='tilename',
            mode='replace',
            selector='.someselector'
        )
        event = AjaxEvent(
            target='http://example.com',
            name='contextchanged',
            selector='.contextsensitiv',
            data=dict(key='value')
        )
        message = AjaxMessage(
            payload='Some info message',
            flavor='info',
            selector='None'
        )
        overlay = AjaxOverlay(
            selector='#ajax-overlay',
            action='someaction',
            target='http://example.com',
            close=False,
            content_selector='.overlay_content',
            css='additional-css-class'
        )
        path = AjaxPath(
            'foo/bar',
            target='http://example.com/foo/bar',
            action='layout:#layout:replace',
            event='contextchanged:#someid',
            overlay='acionname:#custom-overlay:.custom_overlay_content',
            overlay_css='additional-overlay-css-class'
        )

        continuation = [action, event, message, overlay, path]
        afc = AjaxFormContinue(result, continuation)
        self.assertEqual(afc.form, '')

        afc_next = json.loads(afc.next)
        self.assertEqual(afc_next, [{
            "mode": "replace",
            "selector": ".someselector",
            "type": "action",
            "target": "http://example.com",
            "name": "tilename"
        }, {
            "selector": ".contextsensitiv",
            "type": "event",
            "target": "http://example.com",
            "name": "contextchanged",
            "data": {
                "key": "value"
            }
        }, {
            "flavor": "info",
            "type": "message",
            "payload": "Some info message",
            "selector": "None"
        }, {
            "target": "http://example.com",
            "content_selector": ".overlay_content",
            "selector": "#ajax-overlay",
            "action": "someaction",
            "close": False,
            "type": "overlay",
            "css": "additional-css-class"
        }, {
            "overlay_css": "additional-overlay-css-class",
            "target": "http://example.com/foo/bar",
            "overlay": "acionname:#custom-overlay:.custom_overlay_content",
            "action": "layout:#layout:replace",
            "path": "foo/bar",
            "type": "path",
            "event": "contextchanged:#someid"
        }])

    def test_render_ajax_form(self):
        self.assertEqual(ajax_form_template.split('\n'), [
            '<div id="ajaxform">',
            '    %(form)s',
            '</div>',
            '<script language="javascript" type="text/javascript">',
            "    var container = document.getElementById('ajaxform');",
            '    var child = container.firstChild;',
            '    while(child != null && child.nodeType == 3) {',
            '        child = child.nextSibling;',
            '    }',
            "    parent.bdajax.render_ajax_form(child, '%(selector)s', '%(mode)s', %(next)s);",
            '</script>',
            ''
        ])

        # Provide a dummy Form
        with self.layer.hook_tile_reg():
            @tile(name='ajaxtestform')
            class AjaxTestForm(Form):
                def prepare(self):
                    self.form = factory(
                        'form',
                        name='ajaxtestform',
                        props={
                            'action': 'http://example.com/foo',
                        })
                    self.form['foo'] = factory(
                        'field:error:text',
                        props={
                            'required': 1,
                        })
                    self.form['save'] = factory(
                        'submit',
                        props={
                            'action': 'save',
                            'expression': True,
                            'handler': self.save,
                            'next': self.next,
                            'label': 'Save',
                        })

                def save(self, widget, data):
                    pass

                def next(self, request):
                    url = 'http://example.com'
                    return [
                        AjaxAction(url, 'content', 'inner', '#content'),
                        AjaxEvent(url, 'contextchanged', '.contextsensitiv')
                    ]

        # Test unauthorized
        root = get_root()
        request = self.layer.new_request()
        res = render_ajax_form(root, request, 'ajaxtestform')
        self.checkOutput("""
        <div id="ajaxform">\n    \n</div>\n<script language="javascript"
        ...HTTPForbidden: Unauthorized: tile <...AjaxTestForm object at ...>
        failed permission check...
        """, res.text)

        # Test authorized with form extraction failure
        with self.layer.authenticated('max'):
            request.params['ajax'] = '1'
            request.params['ajaxtestform.foo'] = ''
            request.params['action.ajaxtestform.save'] = 1
            response = render_ajax_form(root, request, 'ajaxtestform')
            result = str(response)

        self.assertTrue(result.find('<div class="errormessage">') != -1)
        self.assertTrue(result.find('<script language="javascript"') != -1)
        self.assertTrue(result.find('parent.bdajax.render_ajax_form(child, ') != -1)

        # Test with form processing passing
        with self.layer.authenticated('max'):
            request.params['ajaxtestform.foo'] = 'foo'
            response = render_ajax_form(root, request, 'ajaxtestform')
            result = str(response)

        expected = 'parent.bdajax.render_ajax_form(child, \'#content\', \'inner\', [{'
        self.assertTrue(result.find(expected) != -1)

    def test_livesearch(self):
        # Cone provides a livesearch view, but no referring ``ILiveSearch``
        # implementing adapter for it
        root = get_root()
        request = self.layer.new_request()
        request.params['term'] = 'foo'
        self.assertEqual(livesearch(root, request), [])

        # Provide dummy adapter
        @implementer(ILiveSearch)
        @adapter(Interface)
        class LiveSearch(object):
            def __init__(self, model):
                self.model = model

            def search(self, request, query):
                return [{'value': 'Value'}]

        registry = request.registry
        registry.registerAdapter(LiveSearch)

        self.assertEqual(livesearch(root, request), [{'value': 'Value'}])



from cone.app.browser.ajax import ajax_continue
from cone.app.browser.ajax import ajax_message
from cone.app.browser.ajax import AjaxEvent
from cone.app.browser.utils import make_query
from cone.app.browser.utils import make_url
from cone.tile import Tile
from cone.tile import tile
from node.interfaces import IOrder
from pyramid.i18n import get_localizer
from pyramid.i18n import TranslationStringFactory


_ = TranslationStringFactory('cone.app')


class MoveAction(Tile):

    def move(self):
        raise NotImplementedError(
            'Abstract ``MoveAction`` does not implement ``move``'
        )

    def continuation(self, url):
        return [AjaxEvent(url, 'contextchanged', '#layout')]

    def show_error(self, message):
        localizer = get_localizer(self.request)
        ajax_message(self.request, localizer.translate(message), 'error')

    def render(self):
        model = self.model
        parent = model.parent
        if not IOrder.providedBy(parent):
            title = model.metadata.get('title', model.name)
            message = _(
                'object_not_movable',
                default='Object "${title}" not movable',
                mapping={'title': title}
            )
            self.show_error(message)
            return u''
        if (
            not parent.properties.action_move
            or not self.request.has_permission('change_order', parent)
        ):
            message = _(
                'object_moving_not_permitted',
                default='You are not permitted to move this object'
            )
            self.show_error(message)
            return u''
        self.move()
        parent()
        query = make_query(
            contenttile='listing',
            b_page=self.request.params.get('b_page'),
            size=self.request.params.get('size')
        )
        url = make_url(self.request, node=parent, query=query)
        ajax_continue(self.request, self.continuation(url))
        return u''


@tile(name='move_up', permission='view')
class MoveUpAction(MoveAction):

    def move(self):
        model = self.model
        parent = model.parent
        parent.swap(model, parent[parent.prev_key(model.name)])


@tile(name='move_down', permission='view')
class MoveDownAction(MoveAction):

    def move(self):
        model = self.model
        parent = model.parent
        parent.swap(model, parent[parent.next_key(model.name)])



from cone.app import testing
from cone.app.browser.actions import ViewLink
from cone.app.browser.table import RowData
from cone.app.browser.table import Table
from cone.app.browser.table import TableBatch
from cone.app.browser.table import TableSlice
from cone.app.model import BaseNode
from cone.tile import render_tile
from cone.tile import tile
from cone.tile.tests import TileTestCase
from datetime import datetime
from pyramid.httpexceptions import HTTPForbidden


class DummyTable(Table):
    default_slicesize = 10

    @property
    def item_count(self):
        # ``item_count`` is used to calculate the slice
        return 21

    def sorted_rows(self, start, end, sort, order):
        # ``sorted_rows`` is responsible to generate the table rows data. It
        # gets passed ``start``, ``end``, ``sort`` and ``order`` and must
        # return a list of ``RowData`` instances
        rows = []
        for i in range(self.item_count):
            row_data = RowData()
            row_data['col_1'] = 'Col 1 Value'
            row_data['col_2'] = 'Col 2 Value'
            rows.append(row_data)
        return rows[start:end]


class TestBrowserTable(TileTestCase):
    layer = testing.security

    def test_Table(self):
        table = Table('cone.app:browser/templates/table.pt', None, 'table')
        err = self.expectError(
            NotImplementedError,
            lambda: table.item_count
        )
        expected = 'Abstract table does not implement ``item_count``.'
        self.assertEqual(str(err), expected)

        err = self.expectError(
            NotImplementedError,
            table.sorted_rows,
            None,
            None,
            None,
            None
        )
        expected = 'Abstract table does not implement ``sorted_rows``.'
        self.assertEqual(str(err), expected)

    def test_TableBatch(self):
        model = BaseNode()
        request = self.layer.new_request()

        table = DummyTable(
            'cone.app:browser/templates/table.pt',
            None,
            'table'
        )
        table.request = request
        batch = TableBatch(table)
        batch.model = model
        batch.request = request

        self.assertEqual(batch.vocab, [{
            'current': True,
            'visible': True,
            'url': 'http://example.com/?b_page=0&size=10',
            'page': '1'
        }, {
            'current': False,
            'visible': True,
            'url': 'http://example.com/?b_page=1&size=10',
            'page': '2'
        }, {
            'current': False,
            'visible': True,
            'url': 'http://example.com/?b_page=2&size=10',
            'page': '3'
        }])

    def test_TableSlice(self):
        model = BaseNode()
        request = self.layer.new_request()

        table = DummyTable(
            'cone.app:browser/templates/table.pt',
            None,
            'table'
        )
        table.request = request
        slice = TableSlice(table, model, request)

        self.assertEqual(slice.slice, (0, 10))
        self.assertEqual(slice.rows[0]['col_1'], 'Col 1 Value')
        self.assertEqual(slice.rows[0]['col_2'], 'Col 2 Value')

    def test_col_defs(self):
        model = BaseNode()
        request = self.layer.new_request()

        table = DummyTable(
            'cone.app:browser/templates/table.pt',
            None,
            'table'
        )

        # In order to get row_data rendered inside table, column definitions
        # must be defined
        self.assertEqual(table.col_defs, [])

        res = table(model, request)
        self.assertFalse(res.find('Col 1 Value') > -1)
        self.assertFalse(res.find('Col 2 Value') > -1)

        table.col_defs = [{
            'id': 'col_1',
            'title': 'Col 1',
            'sort_key': 'col_1',
            'sort_title': 'Sort by col 1',
            'content': 'string'
        }, {
            'id': 'col_2',
            'title': 'Col 2',
            'sort_key': 'col_2',
            'sort_title': 'Sort by col 2',
            'content': 'string'
        }]

        res = table(model, request)
        self.assertTrue(res.find('Col 1 Value') > -1)
        self.assertTrue(res.find('Col 2 Value') > -1)

    def test_table_tile(self):
        with self.layer.hook_tile_reg():
            @tile(name='mytabletile',
                  path='cone.app:browser/templates/table.pt',
                  permission='view')
            class MyTable(Table):
                table_id = 'mytable'
                table_css = 'mytable'
                table_tile_name = 'mytabletile'
                col_defs = [{
                    'id': 'col_1',
                    'title': 'Col 1',
                    'sort_key': None,
                    'sort_title': None,
                    'content': 'structure'
                }, {
                    'id': 'col_2',
                    'title': 'Col 2',
                    'sort_key': 'col_2',
                    'sort_title': 'Sort by col 2',
                    'content': 'string'
                }, {
                    'id': 'col_3',
                    'title': 'Col 3',
                    'sort_key': 'col_3',
                    'sort_title': 'Sort by col 3',
                    'content': 'datetime'
                }]
                default_sort = 'col_2'
                default_order = 'desc'
                default_slicesize = 10
                query_whitelist = ['foo']  # additional query params to consider

                @property
                def item_count(self):
                    return 20

                def sorted_rows(self, start, end, sort, order):
                    rows = []
                    for i in range(self.item_count):
                        row_data = RowData()
                        # structure
                        row_data['col_1'] = ViewLink()(self.model, self.request)
                        # string
                        row_data['col_2'] = 'Col 2 -> %i' % i
                        # datetime value
                        row_data['col_3'] = datetime(2011, 4, 1)
                        # append row data
                        rows.append(row_data)
                    # sorting goes here (i.e.)
                    return rows[start:end]

        model = BaseNode()
        model.properties.action_view = True
        model.metadata.title = 'Foo'
        request = self.layer.new_request()

        # Rendering fails unauthorized, 'view' permission is required
        err = self.expectError(
            HTTPForbidden,
            render_tile,
            model,
            request,
            'mytabletile'
        )
        self.checkOutput("""
        Unauthorized: tile <...MyTable object at ...> failed permission check
        """, str(err))

        # Render authenticated
        request.params['foo'] = 'bar'
        with self.layer.authenticated('max'):
            rendered = render_tile(model, request, 'mytabletile')
        expected = '<div id="mytable"'
        self.assertTrue(rendered.find(expected) > -1)

        expected = 'panel-default mytable"'
        self.assertTrue(rendered.find(expected) > -1)

        # Sort header with query white list param
        expected = (
            'ajax:target="http://example.com/'
            '?b_page=1&amp;foo=bar&amp;order=desc&amp;size=10&amp;sort=col_2"'
        )

        self.assertTrue(rendered.find(expected) > -1)

        # Structure content
        self.checkOutput("""
        ...<a
        id="toolbaraction-view"
        href="http://example.com/"
        ajax:bind="click"
        ajax:target="http://example.com/"
        ajax:action="content:#content:inner"
        ajax:path="href"
        >&nbsp;Foo</a>...
        """, rendered)

        # String
        expected = 'Col 2 -&gt; 1'
        self.assertTrue(rendered.find(expected) > -1)

        # Datetime
        expected = '01.04.2011 00:00'
        self.assertTrue(rendered.find(expected) > -1)



from cone.app import get_root
from cone.app import register_plugin_config
from cone.app import testing
from cone.app.browser.ajax import AjaxAction
from cone.app.browser.form import Form
from cone.app.browser.settings import settings_tab_content
from cone.app.browser.settings import SettingsBehavior
from cone.app.model import BaseNode
from cone.tile import render_tile
from cone.tile import Tile
from cone.tile import tile
from cone.tile.tests import TileTestCase
from plumber import plumbing
from yafowil.base import factory


class SomeSettings(BaseNode):
    pass


class OtherSettings(BaseNode):
    pass


class TestBrowserSettings(TileTestCase):
    layer = testing.security

    def test_register_plugin_config(self):
        root = get_root()
        settings = root['settings']
        settings.factories.clear()

        register_plugin_config('foo', SomeSettings)
        register_plugin_config('bar', SomeSettings)
        register_plugin_config('baz', OtherSettings)
        self.assertEqual(settings.factories.keys(), ['foo', 'bar', 'baz'])

        err = self.expectError(
            ValueError,
            register_plugin_config,
            'baz',
            OtherSettings
        )
        self.assertEqual(str(err), "Config with name 'baz' already registered.")

    def test_settings_content_tile(self):
        root = get_root()
        settings = root['settings']
        settings.factories.clear()
        register_plugin_config('foo', SomeSettings)
        register_plugin_config('bar', SomeSettings)
        register_plugin_config('baz', OtherSettings)

        request = self.layer.new_request()

        # Login and render 'content' tile on ``Settings`` node
        with self.layer.authenticated('manager'):
            res = render_tile(settings, request, 'content')
        self.assertTrue(res.find('foo</a>') > -1)
        self.assertTrue(res.find('bar</a>') > -1)
        self.assertTrue(res.find('baz</a>') > -1)

        # 'content' tile for ``SomeSettings``
        with self.layer.hook_tile_reg():
            @tile(name='content', interface=SomeSettings)
            class SomeSettingsTile(Tile):
                def render(self):
                    return '<div>Settings Contents</div>'

        with self.layer.authenticated('manager'):
            res = render_tile(settings['foo'], request, 'content')
        self.assertEqual(res, '<div>Settings Contents</div>')

        # 'content' tile for ``OtherSettings`` which raises an exception at
        # render time
        with self.layer.hook_tile_reg():
            @tile(name='content', interface=OtherSettings)
            class OtherSettingsTile(Tile):
                def render(self):
                    msg = 'This tile can not be rendered for some reason'
                    raise Exception(msg)

        with self.layer.authenticated('manager'):
            err = self.expectError(
                Exception,
                render_tile,
                settings['baz'],
                request,
                'content'
            )
        self.assertEqual(
            str(err),
            'This tile can not be rendered for some reason'
        )

    def test_SettingsBehavior(self):
        root = get_root()
        settings = root['settings']
        settings.factories.clear()
        register_plugin_config('foo', SomeSettings)

        with self.layer.hook_tile_reg():
            @tile(name='editform', interface=SomeSettings)
            @plumbing(SettingsBehavior)
            class SomeSettingsForm(Form):
                def prepare(self):
                    form = factory(
                        u'form',
                        name='editform',
                        props={
                            'action': self.nodeurl
                        })
                    form['foo'] = factory(
                        'field:label:text',
                        props={
                            'label': 'Foo',
                        })
                    form['save'] = factory(
                        'submit',
                        props={
                            'action': 'save',
                            'expression': True,
                            'handler': None,
                            'next': self.next,
                            'label': 'Save',
                        })
                    self.form = form

        request = self.layer.new_request()
        request.params['action.editform.save'] = '1'
        request.params['editform.foo'] = 'foo'
        request.params['ajax'] = '1'

        with self.layer.authenticated('manager'):
            res = render_tile(settings['foo'], request, 'editform')
        self.assertEqual(res, u'')

        action = request.environ['cone.app.continuation'][0]
        self.assertTrue(isinstance(action, AjaxAction))
        self.assertEqual(action.selector, '.foo')

    def test_settings_tab_content(self):
        root = get_root()
        settings = root['settings']
        settings.factories.clear()
        register_plugin_config('foo', SomeSettings)
        register_plugin_config('baz', OtherSettings)

        with self.layer.hook_tile_reg():
            @tile(name='content', interface=SomeSettings)
            class SomeSettingsTile(Tile):
                def render(self):
                    return '<div>Settings Contents</div>'

            @tile(name='content', interface=OtherSettings)
            class OtherSettingsTile(Tile):
                def render(self):
                    msg = 'This tile can not be rendered for some reason'
                    raise Exception(msg)

        # Ajax view for tabs
        request = self.layer.new_request()
        with self.layer.authenticated('manager'):
            response = settings_tab_content(settings['foo'], request)
        self.assertEqual(
            response.text,
            '<div class="foo"><div>Settings Contents</div></div>'
        )

        with self.layer.authenticated('manager'):
            response = settings_tab_content(settings['baz'], request)
        self.checkOutput("""
        <div class="baz">...Exception: This tile can not be rendered for some
        reason\n</pre></div></div>
        """, response.text)



from cone.app.utils import app_config
from cone.tile import Tile
from cone.tile import tile
from pyramid.view import view_config
from webob import Response
import cone.app
import os
import pkg_resources


def is_remote_resource(resource):
    return resource.startswith('http://') \
        or resource.startswith('https://') \
        or resource.startswith('//')


class MergedAssets(object):

    def __init__(self, request):
        self.request = request
        self.merged_js_assets = cone.app.cfg.merged.js
        self.merged_css_assets = cone.app.cfg.merged.css
        self.merged_print_css_assets = cone.app.cfg.merged.print_css

    def merged_assets(self, assets):
        if self.request.authenticated_userid:
            assets = assets.public + assets.protected
        else:
            assets = assets.public
        data = ''
        for view, subpath in assets:
            path = pkg_resources.resource_filename(
                view.package_name,
                os.path.join(view.docroot, subpath)
            )
            with open(path, 'r') as asset:
                data += asset.read() + '\n\n'
        return data

    @property
    def merged_js(self):
        return self.merged_assets(self.merged_js_assets)

    @property
    def merged_css(self):
        return self.merged_assets(self.merged_css_assets)

    @property
    def merged_print_css(self):
        return self.merged_assets(self.merged_print_css_assets)


@view_config(name='cone.js')
def cone_js(model, request):
    assets = MergedAssets(request)
    response = Response(assets.merged_js)
    response.headers['Content-Type'] = 'application/javascript'
    return response


@view_config(name='cone.css')
def cone_css(model, request):
    assets = MergedAssets(request)
    response = Response(assets.merged_css)
    response.headers['Content-Type'] = 'text/css'
    return response


@view_config(name='print.css')
def print_css(model, request):
    assets = MergedAssets(request)
    response = Response(assets.merged_print_css)
    response.headers['Content-Type'] = 'text/css'
    return response


@tile(name='resources', path='templates/resources.pt', permission='login')
class Resources(Tile):
    """Resources tile.

    XXX: either switch to resource management lib here or use resource
         management middleware.
    """

    @property
    def authenticated(self):
        return self.request.authenticated_userid

    @property
    def js(self):
        return self.resources(app_config().js)

    @property
    def css(self):
        return self.resources(app_config().css)

    def resources(self, reg):
        ret = list()
        for res in reg['public']:
            ret.append(self.resource_url(res))
        if self.authenticated:
            for res in reg['protected']:
                ret.append(self.resource_url(res))
        return ret

    def resource_url(self, resource):
        if is_remote_resource(resource):
            return resource
        return '{}/{}'.format(self.request.application_url, resource)



from cone.app.compat import configparser
from cone.app.compat import IS_PY2
from cone.app.compat import ITER_TYPES
from cone.app.interfaces import IAdapterNode
from cone.app.interfaces import IApplicationEnvironment
from cone.app.interfaces import IApplicationNode
from cone.app.interfaces import ICopySupport
from cone.app.interfaces import IFactoryNode
from cone.app.interfaces import ILayoutConfig
from cone.app.interfaces import ILeafNode
from cone.app.interfaces import IMetadata
from cone.app.interfaces import INodeInfo
from cone.app.interfaces import IProperties
from cone.app.interfaces import ITranslation
from cone.app.interfaces import IUUIDAsName
from cone.app.security import acl_registry
from cone.app.utils import app_config
from cone.app.utils import DatetimeHelper
from node import schema
from node.behaviors import AsAttrAccess
from node.behaviors import Attributes
from node.behaviors import ChildFactory
from node.behaviors import DefaultInit
from node.behaviors import Lifecycle
from node.behaviors import MappingAdopt
from node.behaviors import MappingConstraints
from node.behaviors import MappingNode
from node.behaviors import OdictStorage
from node.behaviors import Schema
from node.behaviors import UUIDAware
from node.behaviors import VolatileStorageInvalidate
from node.interfaces import IOrdered
from node.interfaces import IUUID
from node.interfaces import IUUIDAware
from node.utils import instance_property
from node.utils import safe_decode
from node.utils import safe_encode
from odict import odict
from plumber import Behavior
from plumber import default
from plumber import finalize
from plumber import override
from plumber import plumbing
from pyramid.i18n import negotiate_locale_name
from pyramid.i18n import TranslationStringFactory
from pyramid.security import ALL_PERMISSIONS
from pyramid.security import Allow
from pyramid.security import Deny
from pyramid.security import Everyone
from pyramid.threadlocal import get_current_registry
from pyramid.threadlocal import get_current_request
from zope.interface import implementer
import copy
import logging
import os
import uuid


logger = logging.getLogger('cone.app')

try:
    from lxml import etree
except ImportError:  # pragma: no cover
    logger.warning(
        '``lxml`` not present. '
        '``cone.app.model.XMLProperties`` will not work')


_ = TranslationStringFactory('cone.app')


_node_info_registry = dict()


def register_node_info(name, info):
    _node_info_registry[name] = info


# B/C removed as of cone.app 1.1
registerNodeInfo = register_node_info


def get_node_info(name):
    if name in _node_info_registry:
        return _node_info_registry[name]


# B/C removed as of cone.app 1.1
getNodeInfo = get_node_info


class node_info(object):
    """Node info decorator.
    """

    def __init__(self, name, title=None, description=None,
                 factory=None, icon=None, addables=[]):
        self.name = name
        self.title = title
        self.description = description
        self.factory = factory
        self.icon = icon
        self.addables = addables

    def __call__(self, cls):
        cls.node_info_name = self.name
        info = NodeInfo()
        info.node = cls
        info.title = self.title
        info.description = self.description
        info.factory = self.factory
        info.addables = self.addables
        info.icon = self.icon
        register_node_info(cls.node_info_name, info)
        return cls


@implementer(IApplicationEnvironment)
class AppEnvironment(Behavior):

    @default
    @property
    def request(self):
        return get_current_request()

    @default
    @property
    def registry(self):
        return get_current_registry()


@implementer(IApplicationNode)
class AppNode(Behavior):
    node_info_name = default('')

    @default
    @property
    def __acl__(self):
        return acl_registry.lookup(self.__class__, self.node_info_name)

    @default
    @instance_property
    def properties(self):
        props = Properties()
        props.in_navtree = False
        return props

    @default
    @instance_property
    def metadata(self):
        name = self.name
        if not name:
            name = _('no_title', default='No Title')
        metadata = Metadata()
        metadata.title = name
        return metadata

    @default
    @property
    def nodeinfo(self):
        info = get_node_info(self.node_info_name)
        if not info:
            info = NodeInfo()
            info.title = str(self.__class__)
            info.node = self.__class__
            info.icon = app_config().default_node_icon
        return info


@implementer(ILeafNode)
class LeafNode(AppNode):

    @default
    def __getitem__(self, name):
        raise KeyError(name)

    @default
    def __setitem__(self, name, value):
        raise KeyError(name)

    @default
    def __delitem__(self, name):
        raise KeyError(name)

    @default
    def __iter__(self):
        return iter([])


@plumbing(
    AppNode,
    AsAttrAccess,
    MappingConstraints,
    MappingAdopt,
    Attributes,
    DefaultInit,
    MappingNode,
    Lifecycle,
    OdictStorage)
class BaseNode(object):
    pass


@implementer(IFactoryNode)
@plumbing(
    VolatileStorageInvalidate,
    ChildFactory)
class FactoryNode(BaseNode):
    pass


class AppRoot(FactoryNode):
    """Application root.
    """
    factories = odict()

    @instance_property
    def properties(self):
        return Properties()

    @instance_property
    def metadata(self):
        return Metadata()


class AppSettings(FactoryNode):
    """Applications Settings container.
    """
    __acl__ = [
        (Allow, 'role:manager', ['view', 'manage']),
        (Allow, Everyone, 'login'),
        (Deny, Everyone, ALL_PERMISSIONS),
    ]
    factories = odict()

    @instance_property
    def properties(self):
        props = Properties()
        props.in_navtree = False
        props.skip_mainmenu = True
        props.icon = 'ion-ios7-gear'
        return props

    @instance_property
    def metadata(self):
        metadata = Metadata()
        metadata.title = _('settings', default='Settings')
        return metadata


@implementer(IAdapterNode)
class AdapterNode(BaseNode):

    def __init__(self, model, name=None, parent=None):
        BaseNode.__init__(self, name=name, parent=parent)
        self.model = model

    def __iter__(self):
        for key in self.model:
            yield key

    iterkeys = __iter__

    @property
    def attrs(self):
        return self.model.attrs


@implementer(IUUID)
class NamespaceUUID(Behavior):
    """Behavior calculating ``uuid`` by node path and namespace.
    """
    uuid_namespace = default(uuid.UUID('83438507-fdff-45a2-af47-1e001884eab9'))

    @property
    def uuid(self):
        if self.__name__ is None and self.__parent__ is None:
            return None
        return uuid.uuid5(
            self.uuid_namespace,
            '/'.join([_ for _ in self.path if _ is not None])
        )

    @finalize
    @uuid.setter
    def uuid(self, uuid):
        msg = 'Ignore attempt to set {}.uuid'.format(self.__class__.__name__)
        raise NotImplementedError(msg)


class UUIDAttributeAware(UUIDAware):
    """UUIDAware deriving behavior storing the uid on node attributes.
    """
    uuid_attribute_name = default('uuid')

    @property
    def uuid(self):
        return self.attrs.get(self.uuid_attribute_name)

    @default
    @uuid.setter
    def uuid(self, value):
        self.attrs[self.uuid_attribute_name] = value


@implementer(IUUIDAsName)
class UUIDAsName(UUIDAware):

    @property
    def __name__(self):
        return str(self.uuid)

    @finalize
    @__name__.setter
    def __name__(self, name):
        pass

    @finalize
    def set_uuid_for(self, node, override=False, recursiv=False):
        if IUUIDAware.providedBy(node):
            if override or not node.uuid:
                node.uuid = self.uuid_factory()
        if recursiv:
            for k, v in node.items():
                self.set_uuid_for(v, override, recursiv)
                if IUUIDAware.providedBy(v) and override:
                    if IOrdered.providedBy(v):
                        # XXX: improve
                        node.storage.alter_key(k, v.name)
                    else:
                        del node[k]
                        node[v.name] = v


@implementer(ICopySupport)
class CopySupport(Behavior):
    """Plumbing behavior for copy support.
    """
    supports_cut = default(True)
    supports_copy = default(True)
    supports_paste = default(True)


class LanguageSchema:

    def __iter__(self):
        from cone.app import cfg
        return iter(cfg.available_languages)

    def __contains__(self, key):
        return key in iter(self)

    def __getitem__(self, key):
        if key in self:
            return schema.Str()
        raise KeyError(key)

    def get(self, key, default=None):
        if key in self:
            return self[key]
        return default

    def keys(self):
        return list(iter(self))


@implementer(ITranslation)
class Translation(Schema, AppEnvironment):
    schema = override(LanguageSchema())

    @default
    @property
    def value(self):
        request = self.request
        if request:
            lang = negotiate_locale_name(request)
        else:
            settings = self.registry.settings or {}
            lang = settings.get('default_locale_name', 'en')
        value = self.get(lang)
        if not value:
            value = self.name
        return value

    @default
    @property
    def canonical_value(self):
        settings = self.registry.settings or {}
        lang = settings.get('default_locale_name', 'en')
        value = self.get(lang)
        if not value:
            value = self.name
        return value


o_getattr = object.__getattribute__
o_setattr = object.__setattr__


@implementer(IProperties)
class Properties(object):
    # XXX: extend by schema

    def __init__(self, data=None):
        if data is None:
            data = dict()
        o_setattr(self, '_data', data)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __getitem__(self, key):
        return o_getattr(self, '_data')[key]

    def get(self, key, default=None):
        return o_getattr(self, '_data').get(key, default)

    def __contains__(self, key):
        return key in o_getattr(self, '_data')

    def __getattr__(self, name):
        if name == '__provides__':
            return self.__getattribute__(name)
        return o_getattr(self, '_data').get(name)

    def __setattr__(self, name, value):
        o_getattr(self, '_data')[name] = value

    def keys(self):
        return o_getattr(self, '_data').keys()

    def __copy__(self):
        return self.__class__(
            data=copy.copy(o_getattr(self, '_data'))
        )

    def __deepcopy__(self, memo):
        return self.__class__(
            data=copy.deepcopy(o_getattr(self, '_data'), memo)
        )


class ProtectedProperties(Properties):

    def __init__(self, context, permissions, data=None):
        """
        >>> properties = ProtectedProperties(
        ...     context,
        ...     permissions={
        ...         'propname': ['permission1', 'permission2']
        ...     })
        """
        super(ProtectedProperties, self).__init__(data=data)
        o_setattr(self, '_context', context)
        o_setattr(self, '_permissions', permissions)

    def _permits(self, property):
        context = o_getattr(self, '_context')
        permissions = o_getattr(self, '_permissions')
        required = permissions.get(property)
        if not required:
            # no security check
            return True
        request = get_current_request()
        for permission in required:
            if request.has_permission(permission, context):
                return True
        return False

    def __getitem__(self, key):
        if not self._permits(key):
            raise KeyError(u"No permission to access '%s'" % key)
        return super(ProtectedProperties, self).__getitem__(key)

    def get(self, key, default=None):
        if not self._permits(key):
            return default
        return super(ProtectedProperties, self).get(key, default)

    def __contains__(self, key):
        if not self._permits(key):
            return False
        return super(ProtectedProperties, self).__contains__(key)

    def __getattr__(self, name):
        if name == '__provides__':
            return self.__getattribute__(name)
        if not self._permits(name):
            return None
        return super(ProtectedProperties, self).get(name)

    def keys(self):
        keys = super(ProtectedProperties, self).keys()
        keys = [key for key in keys if self._permits(key)]
        return keys

    def __copy__(self):
        return self.__class__(
            context=copy.copy(o_getattr(self, '_context')),
            permissions=copy.copy(o_getattr(self, '_permissions')),
            data=copy.copy(o_getattr(self, '_data'))
        )

    def __deepcopy__(self, memo):
        return self.__class__(
            context=copy.copy(o_getattr(self, '_context')),
            permissions=copy.deepcopy(o_getattr(self, '_permissions'), memo),
            data=copy.deepcopy(o_getattr(self, '_data'), memo)
        )


@implementer(ILayoutConfig)
class LayoutConfig(Properties):

    def __init__(self, model=None, request=None):
        super(LayoutConfig, self).__init__(data=None)
        self.model = model
        self.request = request


# B/C, removed as of cone.app 1.1
Layout = LayoutConfig


@implementer(IMetadata)
class Metadata(Properties):
    pass


@implementer(INodeInfo)
class NodeInfo(Properties):
    pass


class XMLProperties(Properties):

    def __init__(self, path, data=None):
        o_setattr(self, '_path', path)
        o_setattr(self, '_data', odict())
        if data:
            o_getattr(self, '_data').update(data)
        self._init()

    def __call__(self):
        file = open(o_getattr(self, '_path'), 'wb')
        file.write(self._xml_repr())
        file.close()

    def __delitem__(self, name):
        data = o_getattr(self, '_data')
        if name in data:
            del data[name]
        else:
            raise KeyError(u"property %s does not exist" % name)

    def _init(self):
        dth = DatetimeHelper()
        path = o_getattr(self, '_path')
        if not path or not os.path.exists(path):
            return
        file = open(path, 'r')
        tree = etree.parse(file)
        file.close()
        root = tree.getroot()
        data = o_getattr(self, '_data')
        for elem in root.getchildren():
            children = elem.getchildren()
            if children:
                subchildren = children[0].getchildren()
                # case dict like
                if subchildren:
                    val = odict()
                    for subelem in children:
                        entry_elems = subelem.getchildren()
                        val[entry_elems[0].text] = entry_elems[1].text
                # case list like
                else:
                    val = list()
                    for subelem in children:
                        value = subelem.text
                        if not value:
                            value = ''
                        val.append(dth.r_value(value.strip()))
                data[elem.tag] = val
            else:
                value = elem.text
                if not value:
                    value = ''
                data[elem.tag] = dth.r_value(value.strip())
        file.close()

    def _xml_repr(self):
        dth = DatetimeHelper()
        root = etree.Element('properties')
        data = o_getattr(self, '_data')
        for key, value in data.items():
            sub = etree.SubElement(root, key)
            if type(value) in ITER_TYPES:
                for item in value:
                    item_elem = etree.SubElement(sub, 'item')
                    item_elem.text = dth.w_value(item)
            elif isinstance(value, dict) or isinstance(value, odict):
                for key, val in value.items():
                    dict_entry_elem = etree.SubElement(sub, 'elem')
                    key_elem = etree.SubElement(dict_entry_elem, 'key')
                    value_elem = etree.SubElement(dict_entry_elem, 'value')
                    key_elem.text = dth.w_value(key)
                    value_elem.text = dth.w_value(val)
            else:
                sub.text = dth.w_value(value)
        return etree.tostring(root, pretty_print=True)

    # testing
    def _keys(self):
        return o_getattr(self, '_data').keys()

    def _values(self):
        return o_getattr(self, '_data').values()

    def __copy__(self):
        return self.__class__(
            path=copy.copy(o_getattr(self, '_path')),
            data=copy.copy(o_getattr(self, '_data'))
        )

    def __deepcopy__(self, memo):
        return self.__class__(
            path=copy.deepcopy(o_getattr(self, '_path'), memo),
            data=copy.deepcopy(o_getattr(self, '_data'), memo)
        )


class ConfigProperties(Properties):
    properties_section = 'properties'
    encoding = 'utf-8'

    def __init__(self, path, data=None):
        o_setattr(self, '_path', path)
        o_setattr(self, '_data', dict())
        if data:
            o_getattr(self, '_data').update(data)
        self._init()

    def __call__(self):
        path = o_getattr(self, '_path')
        config = self.config()
        mode = 'wb' if IS_PY2 else 'w'
        with open(path, mode) as configfile:
            config.write(configfile)

    def __getitem__(self, key):
        try:
            value = self.config().get(self.properties_section, key)
            return safe_decode(value, encoding=self.encoding)
        except configparser.NoOptionError:
            raise KeyError(key)

    def get(self, key, default=None):
        try:
            value = self.config().get(self.properties_section, key)
            return safe_decode(value, encoding=self.encoding)
        except configparser.NoOptionError:
            return default

    def __contains__(self, key):
        try:
            self.config().get(self.properties_section, key)
            return True
        except configparser.NoOptionError:
            return False

    def __getattr__(self, name):
        if name == '__provides__':
            return self.__getattribute__(name)
        try:
            value = self.config().get(self.properties_section, name)
            return safe_decode(value, encoding=self.encoding)
        except configparser.NoOptionError:
            return None

    def __setattr__(self, name, value):
        value = safe_encode(value, encoding=self.encoding) if IS_PY2 else str(value)
        self.config().set(self.properties_section, name, value)

    def __delitem__(self, name):
        config = self.config()
        try:
            config.get(self.properties_section, name)
        except configparser.NoOptionError:
            raise KeyError(u"property %s does not exist" % name)
        config.remove_option(self.properties_section, name)

    def config(self):
        try:
            return o_getattr(self, '_config')
        except AttributeError:
            pass
        config = configparser.ConfigParser()
        path = o_getattr(self, '_path')
        if os.path.exists(path):
            config.read(path)
        else:
            config.add_section(self.properties_section)
        o_setattr(self, '_config', config)
        return o_getattr(self, '_config')

    def _init(self):
        data = o_getattr(self, '_data')
        config = self.config()
        for key, value in data.items():
            value = safe_encode(value, encoding=self.encoding) if IS_PY2 else str(value)
            config.set(self.properties_section, key, value)

    def __copy__(self):
        cpy = self.__class__(
            path=copy.copy(o_getattr(self, '_path')),
            data=copy.copy(o_getattr(self, '_data'))
        )
        cfg = copy.copy(o_getattr(self, '_config'))
        o_setattr(cpy, '_config', cfg)
        return cpy

    def __deepcopy__(self, memo):
        cpy = self.__class__(
            path=copy.deepcopy(o_getattr(self, '_path'), memo),
            data=copy.deepcopy(o_getattr(self, '_data'), memo)
        )
        cfg = copy.copy(o_getattr(self, '_config'))
        o_setattr(cpy, '_config', cfg)
        return cpy



from cone.app.browser import render_main_template
from cone.app.browser.ajax import AjaxEvent
from cone.app.browser.ajax import ajax_continue
from cone.app.browser.form import Form
from cone.app.browser.utils import make_url
from cone.app.security import authenticate
from cone.tile import Tile
from cone.tile import tile
from pyramid.i18n import TranslationStringFactory
from pyramid.security import forget
from pyramid.view import view_config
from webob.exc import HTTPFound
from yafowil.base import ExtractionError
from yafowil.base import factory


_ = TranslationStringFactory('cone.app')


@view_config(name='login')
def login_view(model, request):
    return render_main_template(model, request, contenttile='loginform')


@view_config(name='logout')
def logout_view(model, request):
    headers = forget(request)
    location = request.params.get('came_from', request.application_url)
    return HTTPFound(location=location, headers=headers)


@tile(name='logout')
class Logout(Tile):

    def render(self):
        request = self.request
        request.response.headers = forget(request)
        location = request.params.get('came_from', request.application_url)
        ajax_continue(self.request, AjaxEvent(
            target=location,
            name='contextchanged',
            selector='#layout'
        ))
        return u''


@tile(name='loginform', permission='login')
class LoginForm(Form):
    ajax = False

    def prepare(self):
        action = make_url(self.request, node=self.model, resource='login')
        form = factory(
            u'form',
            name='loginform',
            props={
                'action': action,
                'class': 'form-horizontal',
            })
        form['user'] = factory(
            'field:label:div:help:error:text',
            props={
                'required': _('no_username_given', default='No username given'),
                'label': _('username', default='Username'),
                'label.class_add': 'col-sm-2',
                'div.class_add': 'col-sm-5',
            })
        form['password'] = factory(
            'field:label:div:help:*credentials:error:password',
            props={
                'required': _('no_password_given', default='No password given'),
                'label': _('password', default='Password'),
                'label.class_add': 'col-sm-2',
                'div.class_add': 'col-sm-5',
            },
            custom={
                'credentials': ([self.login], [], [], []),
            })
        actions = form['form_actions'] = factory(
            'field:div',
            props={
                'div.class_add': 'col-sm-offset-2 col-sm-5',
                'structural': True,
            })
        actions['login'] = factory(
            'submit',
            props={
                'action': 'login',
                'expression': True,
                'handler': self.noop,
                'next': self.next,
                'label': _('login', default='Login'),
                'div.class_add': 'col-sm-offset-2 col-sm-5',
            })
        self.form = form

    def noop(self, widget, data):
        pass

    def login(self, widget, data):
        login = data.fetch('loginform.user').extracted
        password = data.fetch('loginform.password').extracted
        webob_req = data.request.request
        self.headers = authenticate(webob_req, login, password)
        if not self.headers:
            raise ExtractionError(
                _('invalid_credentials', default='Invalid Credentials'))

    def next(self, request):
        return HTTPFound(location=request.request.application_url,
                         headers=self.headers)



from cone.app import compat
from cone.app.browser import render_main_template
from cone.app.browser.actions import ActionContext
from cone.app.browser.ajax import ajax_continue
from cone.app.browser.ajax import ajax_form_fiddle
from cone.app.browser.ajax import ajax_message
from cone.app.browser.ajax import AjaxEvent
from cone.app.browser.ajax import AjaxOverlay
from cone.app.browser.ajax import AjaxPath
from cone.app.browser.ajax import render_ajax_form
from cone.app.browser.utils import make_query
from cone.app.browser.utils import make_url
from cone.app.model import AdapterNode
from cone.app.model import BaseNode
from cone.app.model import get_node_info
from cone.app.model import Properties
from cone.app.utils import app_config
from cone.app.utils import node_path
from cone.tile import render_template
from cone.tile import render_tile
from cone.tile import Tile
from cone.tile import tile
from plumber import Behavior
from plumber import default
from plumber import override
from plumber import plumb
from pyramid.i18n import get_localizer
from pyramid.i18n import TranslationStringFactory
from pyramid.view import view_config
from webob.exc import HTTPFound
from yafowil.base import factory
import logging


logger = logging.getLogger('cone.app')
_ = TranslationStringFactory('cone.app')


###############################################################################
# general
###############################################################################

def is_ajax(request):
    return bool(request.params.get('ajax'))


def render_form(model, request, tilename):
    """If form is invoked without hidden ajax field, the main template is
    rendered with tile ``tilename`` as content tile, otherwise
    ``render_ajax_form`` is called, which renders the tile wrapped by some
    javascript calls into a script tag. The ajax response will be rendered into
    the hidden iframe on client side, where ajax continuation is processed.

    XXX: move to cone.app.browser.form
    """
    if is_ajax(request):
        ActionContext(model, request, tilename)
        return render_ajax_form(model, request, tilename)
    return render_main_template(model, request, contenttile=tilename)


class _FormRenderingTile(Tile):
    form_tile_name = ''

    def render(self):
        return render_tile(self.model, self.request, self.form_tile_name)


###############################################################################
# form continuation control
###############################################################################

class CameFromNext(Behavior):
    """Form behavior for form tiles considering ``came_from`` parameter on
    request for form continuation.
    """

    default_came_from = default(None)
    """Default ``came_from`` value considered in ``next`` if no ``came_from``
    parameter given on request.
    """

    write_history_on_next = default(False)
    """Flag whether to write browser history on ``next`` via ``AjaxPath``
    continuation if ajax request.
    """

    @plumb
    def prepare(_next, self):
        """Hook after prepare and set ``came_from`` as proxy field to
        ``self.form``.
        """
        _next(self)
        # read came_from from request
        came_from = self.request.params.get('came_from')
        # fall back to default_came_from if came_from not passed on request
        if came_from is None:
            came_from = self.default_came_from
        # came_from might be None, set to empty string
        if not came_from:
            came_from = ''
        self.form['came_from'] = factory('proxy', value=came_from)

    @default
    def next(self, request):
        """Read ``came_from`` parameter from request and compute next URL.

        If ``came_from`` not found on request, ``default_came_from`` property
        is used.

        If ``came_from`` is special value ``parent``, URL of model parent is
        computed.

        If ``came_from`` is set, it is considered as URL to use. The given URL
        must match the basic application URL, otherwise an error gets logged
        and URL of current model is computed.

        If ``came_from`` is set to empty value, URL of current model is
        computed.
        """
        # read came_from from request
        came_from = request.get('came_from')
        # fall back to default_came_from if came_from not passed on request
        if came_from is None:
            came_from = self.default_came_from
        # use model URL and path if no came_from
        if not came_from:
            url = make_url(request.request, node=self.model)
            path = '/'.join(node_path(self.model))
        # use model parent URL and path if came_from is 'parent'
        elif came_from == 'parent':
            url = make_url(request.request, node=self.model.parent)
            path = '/'.join(node_path(self.model.parent))
        # consider came_from a URL
        else:
            url = compat.unquote(came_from)
            parsed = compat.urlparse.urlparse(url)
            app_loc = compat.urlparse.urlparse(
                self.request.application_url
            ).netloc
            # behave as if no came_from given if application location not
            # matches came_from location
            if app_loc != parsed.netloc:
                logger.error((
                    'CameFromNext.next(): Application location "{}" does not '
                    'match came_from location "{}". Use model for URL '
                    'computing instead'
                ).format(app_loc, parsed.netloc))
                url = make_url(request.request, node=self.model)
                path = '/'.join(node_path(self.model))
            # include query to path
            elif parsed.query:
                path = '{}?{}'.format(parsed.path, parsed.query)
            # query without path
            else:
                path = '{}'.format(parsed.path)
        # ajax continuation definitions if ajax request
        if self.ajax_request:
            event = AjaxEvent(url, 'contextchanged', '#layout')
            # return continuation path and event if browser history should be
            # written
            if self.write_history_on_next:
                cpath = AjaxPath(
                    path,
                    target=url,
                    event='contextchanged:#layout'
                )
                return [cpath, event]
            # return event only if writing browser history should be skipped
            return [event]
        # regular redirection if no ajax request
        return HTTPFound(location=url)


###############################################################################
# form heading
###############################################################################

class FormHeading(Behavior):

    @default
    @property
    def form_heading(self):
        raise NotImplementedError(u'Abstract ``FormHeading`` does not '
                                  u'implement ``form_heading``')


###############################################################################
# content area related forms
###############################################################################

class ContentForm(FormHeading):
    """Form behavior rendering to content area.
    """
    show_heading = default(True)
    show_contextmenu = default(True)

    @default
    @property
    def form_heading(self):
        return _('content_form_heading', default='Content Form Heading')

    @default
    @property
    def rendered_contextmenu(self):
        return render_tile(self.model, self.request, 'contextmenu')

    @plumb
    def __call__(_next, self, model, request):
        ajax_form_fiddle(request, '#content', 'inner')
        form = _next(self, model, request)
        if not form:
            form = u''
        self.rendered_form = form
        path = self.path
        if not path:
            path = 'cone.app.browser:templates/content_form.pt'
        return render_template(
            path, request=request, model=model, context=self)


###############################################################################
# overlay forms
###############################################################################

@view_config(name='overlayform', permission='view')
def overlayform(model, request):
    """View for posting overlay forms to.
    """
    return render_form(model, request, 'overlayformtile')


@tile(name='overlayformtile', permission='view')
class OverlayFormTile(_FormRenderingTile):
    """Entry tile for rendering forms in overlays.
    """
    form_tile_name = 'overlayform'


class OverlayForm(Behavior):
    """Form behavior rendering to overlay.
    """
    action_resource = override('overlayform')
    # XXX: change ``overlay_selector`` to #ajax-form
    overlay_selector = override('#ajax-overlay')
    overlay_content_selector = override('.overlay_content')

    @plumb
    def __call__(_next, self, model, request):
        form = _next(self, model, request)
        selector = '%s %s' % (self.overlay_selector,
                              self.overlay_content_selector)
        ajax_form_fiddle(request, selector, 'inner')
        return form

    @default
    def next(self, request):
        return [AjaxOverlay(selector=self.overlay_selector, close=True)]


# B/C
# deprecated: will be removed in cone.app 1.1
OverlayBehavior = OverlayForm


###############################################################################
# adding
###############################################################################

def default_addmodel_factory(parent, nodeinfo):
    """Default addmodel factory.

    The addmodel factory is responsible to create a model suitable for
    rendering addforms refering to node info.

    parent
        The parent in which the new item should be added
    nodeinfo
        The nodeinfo instance
    """
    if AdapterNode in nodeinfo.node.__bases__:
        addmodel = nodeinfo.node(BaseNode(), None, None)
    else:
        addmodel = nodeinfo.node()
    addmodel.__parent__ = parent
    return addmodel


@tile(name='add_dropdown',
      path='templates/add_dropdown.pt',
      permission='add', strict=False)
class AddDropdown(Tile):

    def make_item(self, info_name, info):
        model = self.model
        request = self.request
        props = Properties()
        query = make_query(factory=info_name)
        url = make_url(request, node=model, resource='add', query=query)
        props.url = url
        target = make_url(request, node=model, query=query)
        props.target = target
        props.title = info.title
        icon = info.icon
        if not icon:
            icon = app_config().default_node_icon
        props.icon = icon
        return props

    @property
    def items(self):
        ret = list()
        addables = self.model.nodeinfo.addables
        if not addables:
            return ret
        for addable in addables:
            info = get_node_info(addable)
            if not info:
                continue
            ret.append(self.make_item(addable, info))
        return ret


@view_config(name='add', permission='add')
def add(model, request):
    return render_form(model, request, 'add')


@tile(name='add', permission='add')
class AddTile(_FormRenderingTile):
    """The add tile is responsible to render add forms depending on given
    factory name. Factory information is fetched from NodeInfo implementation
    registered by factory name.
    """
    form_tile_name = 'addform'

    def render(self):
        nodeinfo = self.info
        if not nodeinfo:
            return _('unknown_factory', default='Unknown factory')
        factory = nodeinfo.factory
        if not factory:
            factory = default_addmodel_factory
        addmodel = factory(self.model, nodeinfo)
        return render_tile(addmodel, self.request, self.form_tile_name)

    @property
    def info(self):
        factory = self.request.params.get('factory')
        allowed = self.model.nodeinfo.addables
        if not factory or not allowed or factory not in allowed:
            return None
        return get_node_info(factory)


class AddFactoryProxy(Behavior):
    """Form behavior for add forms hooking the hidden field 'factory' to
    request parameters.
    """

    @plumb
    def prepare(_next, self):
        """Hook after prepare and set 'factory' as proxy field to form.
        """
        _next(self)
        self.form['factory'] = factory(
            'proxy',
            value=self.request.params.get('factory'),
        )


class AddFormHeading(FormHeading):

    @default
    @property
    def form_heading(self):
        localizer = get_localizer(self.request)
        title = localizer.translate(
            get_node_info(self.model.node_info_name).title)
        heading = localizer.translate(
            _('add_form_heading',
              default='Add: ${title}',
              mapping={'title': title}))
        return heading


class ContentAddForm(AddFactoryProxy,
                     AddFormHeading,
                     ContentForm,
                     CameFromNext):
    """Form behavior rendering add form to content area.
    """
    action_resource = override('add')

    @default
    @property
    def rendered_contextmenu(self):
        return render_tile(self.model.parent, self.request, 'contextmenu')


# B/C
# deprecated: will be removed in cone.app 1.1
AddBehavior = ContentAddForm


###############################################################################
# overlay adding
###############################################################################

@view_config(name='overlayadd', permission='add')
def overlayadd(model, request):
    return render_form(model, request, 'overlayadd')


@tile(name='overlayadd', permission='add')
class OverlayAddTile(AddTile):
    form_tile_name = 'overlayaddform'


class OverlayAddForm(OverlayForm,
                     AddFactoryProxy,
                     AddFormHeading):
    """Add form behavior rendering to overlay.
    """
    action_resource = override('overlayadd')


###############################################################################
# editing
###############################################################################

@view_config(name='edit', permission='edit')
def edit(model, request):
    return render_form(model, request, 'edit')


@tile(name='edit', permission='edit')
class EditTile(_FormRenderingTile):
    form_tile_name = 'editform'


class EditFormHeading(FormHeading):

    @default
    @property
    def form_heading(self):
        info = get_node_info(self.model.node_info_name)
        if info is None:
            return _('edit', default='Edit')
        localizer = get_localizer(self.request)
        heading = localizer.translate(
            _('edit_form_heading',
              default='Edit: ${title}',
              mapping={'title': localizer.translate(info.title)}))
        return heading


class ContentEditForm(EditFormHeading,
                      ContentForm,
                      CameFromNext):
    """Form behavior rendering edit form to content area.
    """
    action_resource = override('edit')


# B/C
# deprecated: will be removed in cone.app 1.1
EditBehavior = ContentEditForm


###############################################################################
# overlay editing
###############################################################################

@view_config(name='overlayedit', permission='edit')
def overlayedit(model, request):
    return render_form(model, request, 'overlayedit')


@tile(name='overlayedit', permission='edit')
class OverlayEditTile(_FormRenderingTile):
    form_tile_name = 'overlayeditform'


class OverlayEditForm(OverlayForm,
                      EditFormHeading):
    """Edit form behavior rendering to overlay.
    """
    action_resource = override('overlayedit')


###############################################################################
# deleting
###############################################################################

@tile(name='delete', permission='delete')
class DeleteAction(Tile):
    show_confirm_deleted = True

    def continuation(self, url):
        return [AjaxEvent(url, 'contextchanged', '#layout')]

    def render(self):
        model = self.model
        title = model.metadata.get('title', model.name)
        if not model.properties.action_delete:
            ts = _('object_not_deletable',
                   default='Object "${title}" not deletable',
                   mapping={'title': title})
            localizer = get_localizer(self.request)
            message = localizer.translate(ts)
            ajax_message(self.request, message, 'error')
            return u''
        content_tile = model.properties.action_delete_tile
        if not content_tile:
            content_tile = 'content'
        parent = model.parent
        del parent[model.name]
        if hasattr(parent, '__call__'):
            parent()
        query = make_query(contenttile=content_tile)
        url = make_url(self.request, node=parent, query=query)
        ajax_continue(self.request, self.continuation(url))
        if self.show_confirm_deleted:
            ts = _('deleted_object',
                   default='Deleted: ${title}',
                   mapping={'title': title})
            localizer = get_localizer(self.request)
            message = localizer.translate(ts)
            ajax_message(self.request, message, 'info')
        return u''



from cone.app import testing
from cone.app.browser import set_related_view
from cone.app.browser.batch import Batch
from cone.app.browser.batch import BatchedItems
from cone.app.browser.batch import BatchedItemsBatch
from cone.app.browser.utils import make_query
from cone.app.browser.utils import make_url
from cone.app.model import BaseNode
from cone.app.utils import node_path
from cone.tile import render_tile
from cone.tile import tile
from cone.tile.tests import TileTestCase


class TestBrowserBatch(TileTestCase):
    layer = testing.security

    def test_abstract_batch(self):
        # Abstract batch tile. A deriving class must implement the ``vocab``
        # property, which promises to return a list of dict like objects,
        # providing the following keys:
        # - ``page`` - the displayed page name, normally a number, or a
        #              character.
        # - ``current`` - Flag whether page is current page or not.
        # - ``visible`` - Flag whether page is visible or not.
        # - ``href`` - href attribute URL.
        # - ``target`` - ajax target URL.
        # - ``url`` - Target URL. B/C. Use dedicated ``href`` and ``target``.

        # Instanciate directly, base tests
        batch = Batch(None, 'render', 'batch')
        # The dummy page
        self.assertEqual(sorted(batch.dummypage.items()), [
            ('current', False),
            ('href', ''),
            ('page', ''),
            ('target', ''),
            ('url', ''),
            ('visible', False)
        ])
        # Ellipsis to display if ``batchrange`` exceeds
        self.assertEqual(batch.ellipsis, u'...')
        # By default empty ``vocab``, subclass must override
        self.assertEqual(batch.vocab, [])
        # Batches are displayed by default
        self.assertTrue(batch.display)
        # Default ``batchrange`` is 30 items. Defines how many pages are
        # displayed
        self.assertEqual(batch.batchrange, 8)
        # Test ``currentpage`` returns None if no currentpage
        self.assertTrue(batch.currentpage is None)
        # Test ``firstpage`` returns None if empty vocab
        self.assertTrue(batch.firstpage is None)
        # Test ``lastpage`` returns None if empty vocab
        self.assertTrue(batch.lastpage is None)
        # Test ``prevpage`` returns None if empty vocab
        self.assertTrue(batch.prevpage is None)
        # Test ``nextpage`` returns None if empty vocab
        self.assertTrue(batch.nextpage is None)
        # No left or right ellipsis sind no batchrange exceeding
        self.assertEqual(batch.leftellipsis, u'')
        self.assertEqual(batch.rightellipsis, u'')
        # No ``pages`` returned for empty vocab
        self.assertEqual(batch.pages, [])

        # Test edge case ``_position_of_current_in_vocab``
        class BuggyBatch(Batch):
            @property
            def currentpage(self):
                return self.dummypage
        self.assertEqual(BuggyBatch()._position_of_current_in_vocab, -1)

    def test_batch(self):
        # Dummy Batch implementation
        class DummyBatch(Batch):
            _vocab = []
            _batchrange = 5

            @property
            def vocab(self):
                return self._vocab

            @property
            def batchrange(self):
                return self._batchrange

        # Test with all pages invisible
        batch = DummyBatch(None, 'render', 'batch')
        for i in range(3):
            batch._vocab.append({
                'current': False,
                'visible': False,
                'page': str(i),
                'href': 'http://example.com/someview',
                'target': 'http://example.com/'
            })

        # If no visible page, ``firstpage`` returns first page from vocab
        self.assertEqual(sorted(batch.firstpage.items()), [
            ('current', False),
            ('href', 'http://example.com/someview'),
            ('page', '0'),
            ('target', 'http://example.com/'),
            ('visible', False)
        ])

        # If no visible page, ``lastpage`` returns last page from vocab
        self.assertEqual(sorted(batch.lastpage.items()), [
            ('current', False),
            ('href', 'http://example.com/someview'),
            ('page', '2'),
            ('target', 'http://example.com/'),
            ('visible', False)
        ])

        # No visible pages in vocab return ``dummypage`` on prevpage and
        # nextpage
        self.assertTrue(batch.prevpage == batch.dummypage)
        self.assertTrue(batch.nextpage == batch.dummypage)

        # Test with visible pages
        batch._vocab = list()
        for i in range(5):
            batch._vocab.append({
                'current': False,
                'visible': True,
                'href': 'http://example.com/someview',
                'target': 'http://example.com/',
                'page': str(i),
            })
        batch._vocab[1]['visible'] = False
        batch._vocab[3]['visible'] = False

        # Set first page current
        batch._vocab[0]['current'] = True

        # First vocab item is visible, ``firstpage`` returns it
        self.assertEqual(sorted(batch.firstpage.items()), [
            ('current', True),
            ('href', 'http://example.com/someview'),
            ('page', '0'),
            ('target', 'http://example.com/'),
            ('visible', True)
        ])

        # Last vocab item is visible, ``lastpage`` returns it
        self.assertEqual(sorted(batch.lastpage.items()), [
            ('current', False),
            ('href', 'http://example.com/someview'),
            ('page', '4'),
            ('target', 'http://example.com/'),
            ('visible', True)
        ])

        # First item is selected, ``prevpage`` returns dummy page
        self.assertEqual(sorted(batch.prevpage.items()), [
            ('current', False),
            ('href', ''),
            ('page', ''),
            ('target', ''),
            ('url', ''),
            ('visible', False)
        ])

        # ``nextpage`` returns next visible page, vocab[1] is skipped
        self.assertEqual(sorted(batch.nextpage.items()), [
            ('current', False),
            ('href', 'http://example.com/someview'),
            ('page', '2'),
            ('target', 'http://example.com/'),
            ('visible', True)
        ])

        # Set last page current
        batch._vocab[0]['current'] = False
        batch._vocab[-1]['current'] = True

        # ``prevpage`` returns next visible page, vocab[3] is skipped
        self.assertEqual(sorted(batch.prevpage.items()), [
            ('current', False),
            ('href', 'http://example.com/someview'),
            ('page', '2'),
            ('target', 'http://example.com/'),
            ('visible', True)
        ])

        # Last item is selected, ``nextpage`` returns dummy page
        self.assertEqual(sorted(batch.nextpage.items()), [
            ('current', False),
            ('href', ''),
            ('page', ''),
            ('target', ''),
            ('url', ''),
            ('visible', False)
        ])

        # Set third page current
        batch._vocab[-1]['current'] = False
        batch._vocab[2]['current'] = True

        # ``prevpage`` returns next visible page, vocab[1] is skipped
        self.assertEqual(sorted(batch.prevpage.items()), [
            ('current', False),
            ('href', 'http://example.com/someview'),
            ('page', '0'),
            ('target', 'http://example.com/'),
            ('visible', True)
        ])

        # ``nextpage`` returns next visible page, vocab[3] is skipped
        self.assertEqual(sorted(batch.nextpage.items()), [
            ('current', False),
            ('href', 'http://example.com/someview'),
            ('page', '4'),
            ('target', 'http://example.com/'),
            ('visible', True)
        ])

        # Inverse visible flags
        batch._vocab[0]['visible'] = False
        batch._vocab[1]['visible'] = True
        batch._vocab[2]['visible'] = False
        batch._vocab[3]['visible'] = True
        batch._vocab[4]['visible'] = False

        # Set second item selected
        batch._vocab[2]['current'] = False
        batch._vocab[1]['current'] = True

        # ``firstpage`` returns first visible page
        self.assertEqual(sorted(batch.firstpage.items()), [
            ('current', True),
            ('href', 'http://example.com/someview'),
            ('page', '1'),
            ('target', 'http://example.com/'),
            ('visible', True)
        ])

        # ``lastpage`` returns last visible page
        self.assertEqual(sorted(batch.lastpage.items()), [
            ('current', False),
            ('href', 'http://example.com/someview'),
            ('page', '3'),
            ('target', 'http://example.com/'),
            ('visible', True)
        ])

        # Selected page is first visible page, ``prevpage`` returns dummypage
        self.assertEqual(sorted(batch.prevpage.items()), [
            ('current', False),
            ('href', ''),
            ('page', ''),
            ('target', ''),
            ('url', ''),
            ('visible', False)
        ])

        # Next visible page
        self.assertEqual(sorted(batch.nextpage.items()), [
            ('current', False),
            ('href', 'http://example.com/someview'),
            ('page', '3'),
            ('target', 'http://example.com/'),
            ('visible', True)
        ])

        # Set fourth item selected
        batch._vocab[1]['current'] = False
        batch._vocab[3]['current'] = True

        # Previous visible page
        self.assertEqual(sorted(batch.prevpage.items()), [
            ('current', False),
            ('href', 'http://example.com/someview'),
            ('page', '1'),
            ('target', 'http://example.com/'),
            ('visible', True)
        ])

        # Selected page is last visible page, ``nextpage`` returns dummypage
        self.assertEqual(sorted(batch.nextpage.items()), [
            ('current', False),
            ('href', ''),
            ('page', ''),
            ('target', ''),
            ('url', ''),
            ('visible', False)
        ])

        # set ``batchrange`` smaller than vocab size
        batch._batchrange = 3
        self.assertEqual(len(batch.pages), 3)

        # Batchrange ends
        self.assertEqual(sorted(batch.pages[0].items()), [
            ('current', False),
            ('href', 'http://example.com/someview'),
            ('page', '2'),
            ('target', 'http://example.com/'),
            ('visible', False)
        ])

        self.assertEqual(sorted(batch.pages[-1].items()), [
            ('current', False),
            ('href', 'http://example.com/someview'),
            ('page', '4'),
            ('target', 'http://example.com/'),
            ('visible', False)
        ])

        self.assertEqual(batch.leftellipsis, u'...')
        self.assertEqual(batch.rightellipsis, u'')

        # Batchrange starts
        batch._vocab[1]['current'] = True
        batch._vocab[3]['current'] = False

        self.assertEqual(sorted(batch.pages[0].items()), [
            ('current', False),
            ('href', 'http://example.com/someview'),
            ('page', '0'),
            ('target', 'http://example.com/'),
            ('visible', False)
        ])

        self.assertEqual(sorted(batch.pages[-1].items()), [
            ('current', False),
            ('href', 'http://example.com/someview'),
            ('page', '2'),
            ('target', 'http://example.com/'),
            ('visible', False)
        ])

        self.assertEqual(batch.leftellipsis, u'')
        self.assertEqual(batch.rightellipsis, u'...')

        # Batchrange between start and end
        batch._vocab[0]['visible'] = True
        batch._vocab[2]['visible'] = True
        batch._vocab[4]['visible'] = True

        batch._vocab[1]['current'] = False
        batch._vocab[2]['current'] = True

        self.assertEqual(sorted(batch.pages[0].items()), [
            ('current', False),
            ('href', 'http://example.com/someview'),
            ('page', '1'),
            ('target', 'http://example.com/'),
            ('visible', True)
        ])
        self.assertEqual(sorted(batch.pages[-1].items()), [
            ('current', False),
            ('href', 'http://example.com/someview'),
            ('page', '3'),
            ('target', 'http://example.com/'),
            ('visible', True)
        ])

        self.assertEqual(batch.leftellipsis, u'...')
        self.assertEqual(batch.rightellipsis, u'...')

    def test_batch_tile(self):
        # Register batch tile
        with self.layer.hook_tile_reg():
            @tile(name='testbatch')
            class TestBatch(Batch):
                @property
                def vocab(self):
                    ret = list()
                    path = node_path(self.model)
                    current = self.request.params.get('b_page', '0')
                    for i in range(10):
                        query = make_query(b_page=str(i))
                        href = make_url(
                            self.request,
                            path=path,
                            resource='someview',
                            query=query
                        )
                        target = make_url(self.request, path=path, query=query)
                        ret.append({
                            'page': '%i' % i,
                            'current': current == str(i),
                            'visible': True,
                            'href': href,
                            'target': target,
                        })
                    return ret

        with self.layer.authenticated('max'):
            model = BaseNode()
            request = self.layer.new_request()
            res = render_tile(model, request, 'testbatch')

        expected = 'href="http://example.com/someview?b_page=1"'
        self.assertTrue(res.find(expected) > -1)
        expected = 'ajax:target="http://example.com/?b_page=1"'
        self.assertTrue(res.find(expected) > -1)
        expected = 'href="http://example.com/someview?b_page=2"'
        self.assertTrue(res.find(expected) > -1)
        expected = 'ajax:target="http://example.com/?b_page=2"'
        self.assertTrue(res.find(expected) > -1)

    def test_bc_batch_tile(self):
        # Test B/C batch vocab rendering::

        with self.layer.hook_tile_reg():
            @tile('bc_testbatch')
            class BCTestBatch(Batch):
                @property
                def vocab(self):
                    ret = list()
                    path = node_path(self.model)
                    current = self.request.params.get('b_page', '0')
                    for i in range(10):
                        query = make_query(b_page=str(i))
                        url = make_url(self.request, path=path, query=query)
                        ret.append({
                            'page': '%i' % i,
                            'current': current == str(i),
                            'visible': True,
                            'url': url
                        })
                    return ret

        with self.layer.authenticated('max'):
            model = BaseNode()
            request = self.layer.new_request()
            res = render_tile(model, request, 'bc_testbatch')

        expected = 'href="http://example.com/?b_page=1"'
        self.assertTrue(res.find(expected) > -1)
        expected = 'ajax:target="http://example.com/?b_page=1"'
        self.assertTrue(res.find(expected) > -1)
        expected = 'href="http://example.com/?b_page=2"'
        self.assertTrue(res.find(expected) > -1)
        expected = 'ajax:target="http://example.com/?b_page=2"'
        self.assertTrue(res.find(expected) > -1)

    def test_abstract_BatchedItems(self):
        # Abstract contracts
        batched_items = BatchedItems()
        batched_items.model = BaseNode()
        batched_items.request = self.layer.new_request()

        err = self.expectError(
            NotImplementedError,
            lambda: batched_items.item_count
        )
        expected = (
            'Abstract ``BatchedItems`` does not implement ``item_count``'
        )
        self.assertEqual(str(err), expected)

        err = self.expectError(
            NotImplementedError,
            lambda: batched_items.slice_items
        )
        expected = (
            'Abstract ``BatchedItems`` does not implement ``items``'
        )
        self.assertEqual(str(err), expected)

        self.assertTrue(batched_items.slice_template is None)

    def test_BatchedItems(self):
        # Concrete ``BatchedItems`` implementation.
        class MyBatchedItems(BatchedItems):
            slice_template = 'cone.app.testing:dummy_batched_items.pt'

            @property
            def item_count(self):
                return len(self.filtered_items)

            @property
            def slice_items(self):
                start, end = self.current_slice
                return self.filtered_items[start:end]

            @property
            def filtered_items(self):
                items = list()
                term = self.filter_term
                term = term.lower() if term else term
                for node in self.model.values():
                    if term and node.name.find(term) == -1:
                        continue
                    items.append(node)
                return items

        # Create model
        model = BaseNode(name='container')
        for i in range(35):
            model['child_{}'.format(i)] = BaseNode()

        # Create batched items with model
        batched_items = MyBatchedItems()
        batched_items.model = model
        batched_items.request = self.layer.new_request()

        # The helper function ``make_query`` considers ``query_whitelist`` and
        # is used for query creation within batched items implementation.
        self.assertEqual(batched_items.query_whitelist, [])

        batched_items.query_whitelist = ['a', 'b']
        batched_items.request.params['a'] = 'a'

        self.assertEqual(batched_items.make_query({'c': 'c'}), '?a=a&b=&c=c')

        # A query parameter which already exists on request gets overwritten
        self.assertEqual(batched_items.make_query({'a': 'b'}), '?a=b&b=')

        # The helper function ``make_url`` uses ``make_query``, thus considers
        # ``query_whitelist`` as well and is used for URL creation within
        # batched items implementation.
        self.assertEqual(
            batched_items.make_url(dict(c='c')),
            u'http://example.com/container?a=a&b=&c=c'
        )

        # It's also possible to pass a model path to ``make_url`` to avoid
        # multiple computing of model path
        path = node_path(model)
        self.assertEqual(
            batched_items.make_url(dict(c='c'), path=path),
            u'http://example.com/container?a=a&b=&c=c'
        )

        # ``BatchedItems`` plumbs ``RelatedViewConsumer`` and considers
        # ``related_view`` if ``include_view`` passed to ``make_url``
        request = batched_items.request = self.layer.new_request()
        set_related_view(request, 'someview')

        self.assertEqual(
            batched_items.make_url(dict(c='c')),
            u'http://example.com/container?a=&b=&c=c'
        )
        self.assertEqual(
            batched_items.make_url(dict(c='c'), include_view=True),
            u'http://example.com/container/someview?a=&b=&c=c'
        )
        self.assertEqual(
            batched_items.make_url(dict(c='c'), path=path),
            u'http://example.com/container?a=&b=&c=c'
        )
        self.assertEqual(
            batched_items.make_url(dict(c='c'), path=path, include_view=True),
            u'http://example.com/container/someview?a=&b=&c=c'
        )

        # Default slice size
        self.assertEqual(batched_items.default_slice_size, 15)
        # Current slice size
        self.assertEqual(batched_items.slice_size, 15)
        # Number of available slice slizes
        self.assertEqual(batched_items.num_slice_sizes, 4)
        # Available slice sizes for slice size selection
        self.assertEqual(batched_items.slice_sizes, [15, 30, 45, 60])

        batched_items.default_slice_size = 10
        batched_items.num_slice_sizes = 5
        self.assertEqual(batched_items.slice_sizes, [10, 20, 30, 40, 50])

        batched_items.default_slice_size = 15
        batched_items.num_slice_sizes = 4

        # Test ``slice_target``
        self.assertEqual(batched_items.query_whitelist, ['a', 'b'])

        request = batched_items.request = self.layer.new_request()
        request.params['a'] = 'a'
        request.params['b'] = 'b'
        request.params['term'] = 'Hello'

        self.assertEqual(batched_items.filter_term, u'Hello')
        self.assertEqual(
            batched_items.slice_target,
            u'http://example.com/container?a=a&b=b&term=Hello'
        )

        # Test ``filter_target``
        self.assertEqual(
            batched_items.filter_target,
            u'http://example.com/container?a=a&b=b&size=15'
        )

        request.params['size'] = '30'
        self.assertEqual(
            batched_items.filter_target,
            u'http://example.com/container?a=a&b=b&size=30'
        )

        # Header template path
        self.assertEqual(
            batched_items.header_template,
            'cone.app.browser:templates/batched_items_header.pt'
        )

        # Rendered header
        self.checkOutput("""
        ...<div class="panel-heading batched_items_header">...
        """, batched_items.rendered_header)

        # Header title. Taken from ``model.metadata`` by default
        self.assertEqual(batched_items.title, 'container')

        # Title can be skipped by setting ``show_title`` to False
        expected = '<span class="label label-primary">container</span>'
        self.assertTrue(batched_items.rendered_header.find(expected) > -1)

        batched_items.show_title = False
        self.assertFalse(batched_items.rendered_header.find(expected) > -1)

        batched_items.show_title = True

        # Slice size can be skipped by setting ``show_slice_size`` to False
        expected = '<select name="size"'
        self.assertTrue(batched_items.rendered_header.find(expected) > -1)

        batched_items.show_slice_size = False
        self.assertFalse(batched_items.rendered_header.find(expected) > -1)

        batched_items.show_slice_size = True

        # CSS class set on slice size selection wrapper
        expected = 'col-xs-4 col-sm3'
        self.assertTrue(batched_items.rendered_header.find(expected) > -1)

        batched_items.slice_size_css = 'col-xs-3 col-sm2'
        self.assertFalse(batched_items.rendered_header.find(expected) > -1)

        batched_items.slice_size_css = 'col-xs-4 col-sm3'

        # Flag whether to show search filter
        expected = '<input name="term"'
        self.assertTrue(batched_items.rendered_header.find(expected) > -1)

        batched_items.show_filter = False
        self.assertFalse(batched_items.rendered_header.find(expected) > -1)

        batched_items.show_filter = True

        # CSS class set on slice search filter
        expected = 'col-xs-3'
        self.assertTrue(batched_items.rendered_header.find(expected) > -1)

        batched_items.filter_css = 'col-xs-4'
        self.assertFalse(batched_items.rendered_header.find(expected) > -1)

        batched_items.filter_css = 'col-xs-3'

        # Additional markup displayed in header
        expected = '<div class="additional">Additional</div>'
        self.assertFalse(batched_items.rendered_header.find(expected) > -1)

        batched_items.head_additional = expected
        self.assertTrue(batched_items.rendered_header.find(expected) > -1)

        batched_items.head_additional = None

        # Batched items pagination. Pagination object is provided by
        # ``pagination`` property on ``BatchedItems``
        request = self.layer.new_request()
        set_related_view(request, 'someview')

        batched_items = MyBatchedItems()
        batched_items.model = BaseNode(name='container')
        batched_items.request = request

        pagination = batched_items.pagination
        self.assertTrue(isinstance(pagination, BatchedItemsBatch))

        pagination.model = batched_items.model
        pagination.request = batched_items.request

        # Pagination batch uses ``page_target`` on ``BatchedItems`` for target
        # URL computing.
        path = node_path(batched_items.model)
        page = '1'
        self.assertEqual(
            batched_items.page_target(path, page),
            u'http://example.com/container?b_page=1&size=15'
        )

        # Pagination batch name is created from batched items ``items_id``
        self.assertEqual(batched_items.items_id, 'batched_items')
        self.assertEqual(pagination.name, 'batched_itemsbatch')

        # Pagination batch only gets displayed if there are batched items.
        self.assertEqual(batched_items.item_count, 0)
        self.assertFalse(pagination.display)
        self.assertEqual(pagination.vocab, [])

        batched_items.model = pagination.model = model
        self.assertEqual(batched_items.item_count, 35)
        self.assertTrue(pagination.display)
        self.assertEqual(batched_items.current_page, 0)

        request.params['b_page'] = '1'
        self.assertEqual(batched_items.current_page, 1)

        vocab = pagination.vocab
        self.assertEqual(len(vocab), 3)

        self.assertEqual(sorted(vocab[0].items()), [
            ('current', False),
            ('href', u'http://example.com/container/someview?b_page=0&size=15'),
            ('page', '1'),
            ('target', u'http://example.com/container?b_page=0&size=15'),
            ('visible', True)
        ])
        self.assertEqual(sorted(vocab[1].items()), [
            ('current', True),
            ('href', u'http://example.com/container/someview?b_page=1&size=15'),
            ('page', '2'),
            ('target', u'http://example.com/container?b_page=1&size=15'),
            ('visible', True)
        ])
        self.assertEqual(sorted(vocab[2].items()), [
            ('current', False),
            ('href', u'http://example.com/container/someview?b_page=2&size=15'),
            ('page', '3'),
            ('target', u'http://example.com/container?b_page=2&size=15'),
            ('visible', True)
        ])

        # Rendered pagination
        self.checkOutput("""
        ...<ul class="pagination pagination-sm">...
        """, batched_items.rendered_pagination)

        # Batched items footer
        batched_items = MyBatchedItems()
        batched_items.model = model
        batched_items.request = self.layer.new_request()

        # Default template path
        self.assertEqual(
            batched_items.footer_template,
            'cone.app.browser:templates/batched_items_footer.pt'
        )
        self.checkOutput("""
        ...<div class="panel-footer batched_items_footer">...
        """, batched_items.rendered_footer)

        # Slice ID
        self.assertEqual(batched_items.slice_id, 'batched_items_slice')

        # Current slice to display as tuple
        self.assertEqual(batched_items.current_slice, (0, 15))

        # Overall item count
        self.assertEqual(batched_items.item_count, 35)

        # Current slice items
        self.checkOutput("""
        [<BaseNode object 'child_0' at ...>,
        ...
        <BaseNode object 'child_14' at ...>]
        """, str(batched_items.slice_items))

        # Chage current page and check again
        request = batched_items.request = self.layer.new_request()
        request.params['b_page'] = '1'
        self.assertEqual(batched_items.current_slice, (15, 30))
        self.checkOutput("""
        [<BaseNode object 'child_15' at ...>,
        ...
        <BaseNode object 'child_29' at ...>]
        """, str(batched_items.slice_items))

        # Change the slice size
        request = batched_items.request = self.layer.new_request()
        request.params['size'] = '10'
        self.assertEqual(batched_items.slice_size, 10)
        self.assertEqual(batched_items.current_slice, (0, 10))
        self.checkOutput("""
        [<BaseNode object 'child_0' at ...>,
        ...
        <BaseNode object 'child_9' at ...>]
        """, str(batched_items.slice_items))

        # Change the filter term
        request = batched_items.request = self.layer.new_request()
        request.params['term'] = '1'
        request.params['size'] = '5'
        self.assertEqual(batched_items.filter_term, u'1')

        self.checkOutput("""
        [<BaseNode object 'child_1' at ...>,
        <BaseNode object 'child_10' at ...>,
        <BaseNode object 'child_11' at ...>,
        <BaseNode object 'child_12' at ...>,
        <BaseNode object 'child_13' at ...>,
        <BaseNode object 'child_14' at ...>,
        <BaseNode object 'child_15' at ...>,
        <BaseNode object 'child_16' at ...>,
        <BaseNode object 'child_17' at ...>,
        <BaseNode object 'child_18' at ...>,
        <BaseNode object 'child_19' at ...>,
        <BaseNode object 'child_21' at ...>,
        <BaseNode object 'child_31' at ...>]
        """, str(batched_items.filtered_items))

        self.assertEqual(batched_items.current_slice, (0, 5))

        self.checkOutput("""
        [<BaseNode object 'child_1' at ...>,
        <BaseNode object 'child_10' at ...>,
        <BaseNode object 'child_11' at ...>,
        <BaseNode object 'child_12' at ...>,
        <BaseNode object 'child_13' at ...>]
        """, str(batched_items.slice_items))

        request.params['b_page'] = '1'
        self.assertEqual(batched_items.current_slice, (5, 10))

        self.checkOutput("""
        [<BaseNode object 'child_14' at ...>,
        <BaseNode object 'child_15' at ...>,
        <BaseNode object 'child_16' at ...>,
        <BaseNode object 'child_17' at ...>,
        <BaseNode object 'child_18' at ...>]
        """, str(batched_items.slice_items))

        # Test ``rendered_slice``
        request = batched_items.request = self.layer.new_request()
        self.checkOutput("""
        <div id="batched_items_slice">
          <div>child_0</div>
          ...
          <div>child_14</div>
        </div>
        """, batched_items.rendered_slice)

        # ``BatchItems`` rendering default template
        self.assertEqual(
            batched_items.path,
            'cone.app.browser:templates/batched_items.pt'
        )

        # Batched items DOM element ID. Used for bdajax binding.
        self.assertEqual(batched_items.items_id, 'batched_items')

        self.checkOutput("""
        ...<div id="batched_items"...
        """, batched_items(model=model, request=self.layer.new_request()))

        batched_items.items_id = 'my_batched_items'

        self.checkOutput("""
        ...<div id="my_batched_items"...
        """, batched_items(model=model, request=self.layer.new_request()))

        batched_items.items_id = 'batched_items'

        # Test ``items_css``
        self.assertEqual(
            batched_items.items_css,
            'batched_items panel panel-default'
        )

        self.checkOutput("""
        ...class="...batched_items ...
        """, batched_items(model=model, request=self.layer.new_request()))

        batched_items.items_css = (
            'my_batched_items batched_items panel panel-default'
        )

        self.checkOutput("""
        ...class="...my_batched_items batched_items ...
        """, batched_items(model=model, request=self.layer.new_request()))

        batched_items.items_css = 'batched_items panel panel-default'

        # Test ``bind_events``
        self.assertEqual(batched_items.bind_events, 'batchclicked')

        self.checkOutput("""
        ...ajax:bind="batchclicked"...
        """, batched_items(model=model, request=self.layer.new_request()))

        # Test ``bind_selectors``
        self.assertEqual(
            batched_items.bind_selectors,
            'batched_itemsbatchsensitiv'
        )

        self.checkOutput("""
        ...class="batched_itemsbatchsensitiv...
        """, batched_items(model=model, request=self.layer.new_request()))

        # Test ``display_header``
        self.assertTrue(batched_items.display_header)

        expected = '<div class="panel-heading batched_items_header">'
        rendered = batched_items(model=model, request=self.layer.new_request())
        self.assertTrue(rendered.find(expected) > -1)

        batched_items.display_header = False
        rendered = batched_items(model=model, request=self.layer.new_request())
        self.assertFalse(rendered.find(expected) > -1)

        batched_items.display_header = True

        # Test ``display_footer``
        self.assertTrue(batched_items.display_header)

        expected = '<div class="panel-footer batched_items_footer">'
        rendered = batched_items(model=model, request=self.layer.new_request())
        self.assertTrue(rendered.find(expected) > -1)

        batched_items.display_footer = False
        rendered = batched_items(model=model, request=self.layer.new_request())
        self.assertFalse(rendered.find(expected) > -1)



from cone.app import testing
from cone.app.browser.utils import authenticated
from cone.app.browser.utils import choose_name
from cone.app.browser.utils import format_date
from cone.app.browser.utils import format_traceback
from cone.app.browser.utils import make_query
from cone.app.browser.utils import make_url
from cone.app.browser.utils import node_icon
from cone.app.browser.utils import request_property
from cone.app.model import BaseNode
from cone.app.model import node_info
from cone.tile.tests import TileTestCase
from datetime import datetime


class TestBrowserUtils(TileTestCase):
    layer = testing.security

    def test_authenticated(self):
        request = self.layer.new_request()
        self.assertEqual(authenticated(request), None)
        with self.layer.authenticated('max'):
            self.assertEqual(authenticated(request), 'max')

    def test_make_query(self):
        self.assertEqual(make_query(foo=None), None)
        self.assertEqual(make_query(foo=[]), None)
        self.assertEqual(make_query(foo='123'), '?foo=123')
        self.assertEqual(make_query(foo=['456', '789']), '?foo=456&foo=789')
        self.assertEqual(make_query(foo=1), '?foo=1')
        self.assertEqual(make_query(foo=1.), '?foo=1.0')
        self.assertEqual(make_query(foo='foo', bar='bar'), '?bar=bar&foo=foo')
        self.assertEqual(
            make_query(
                quote_params=('foo',),
                foo='http://example.com?param=value'
            ),
            '?foo=http%3A//example.com%3Fparam%3Dvalue'
        )
        self.assertEqual(
            make_query(came_from='http://example.com?param=value'),
            '?came_from=http%3A//example.com%3Fparam%3Dvalue'
        )

    def test_make_url(self):
        root = BaseNode()
        root['child'] = BaseNode()
        request = self.layer.new_request()
        self.assertEqual(make_url(request), 'http://example.com/')
        self.assertEqual(
            make_url(request, path=['1', '2', '3']),
            'http://example.com/1/2/3'
        )
        self.assertEqual(
            make_url(request, node=root['child']),
            'http://example.com/child'
        )
        self.assertEqual(
            make_url(request, node=root['child'], resource='foo'),
            'http://example.com/child/foo'
        )
        self.assertEqual(
            make_url(request, node=root['child'], resource='foo', query='&a=1'),
            'http://example.com/child/foo&a=1'
        )

    def test_choose_name(self):
        container = BaseNode()
        container['a'] = BaseNode()
        self.assertEqual(choose_name(container, 'a'), 'a-1')
        container['a-1'] = BaseNode()
        self.assertEqual(choose_name(container, 'a'), 'a-2')
        container['b-1'] = BaseNode()
        self.assertEqual(choose_name(container, 'b'), 'b')

    def test_format_date(self):
        dt = datetime(2011, 3, 15)
        self.assertEqual(format_date(dt), '15.03.2011 00:00')
        self.assertEqual(format_date(dt, long=False), '15.03.2011')
        self.assertEqual(format_date(object()), u'unknown')

    @testing.reset_node_info_registry
    def test_node_icon(self):
        model = BaseNode()
        model.properties.icon = 'my-icon'
        self.assertEqual(node_icon(model), 'my-icon')

        @node_info(
            name='mynode')
        class MyNode(BaseNode):
            pass

        model = MyNode()
        self.assertEqual(
            node_icon(model),
            'glyphicon glyphicon-asterisk'
        )

        @node_info(
            name='othernode',
            icon='other-icon')
        class OtherNode(BaseNode):
            pass

        model = OtherNode()
        self.assertEqual(node_icon(model), 'other-icon')

    def test_request_property(self):
        counter = dict(computed=0)

        class RequestPropertyUsingClass(object):
            def __init__(self, request):
                self.request = request

            @request_property
            def cached_attr(self):
                counter['computed'] += 1
                return 'cached attribute'

        request = self.layer.new_request()
        rpuc = RequestPropertyUsingClass(request)
        self.assertEqual(rpuc.cached_attr, 'cached attribute')
        self.assertEqual(counter['computed'], 1)

        self.assertEqual(rpuc.cached_attr, 'cached attribute')
        self.assertEqual(counter['computed'], 1)

    def test_format_traceback(self):
        class MyException(Exception):
            pass
        try:
            raise MyException('Error!')
        except MyException:
            self.checkOutput("""
            <pre>Traceback (most recent call last):
              File "...", line ..., in test_format_traceback
                raise MyException('Error!')
              ...MyException: Error!
            </pre>
            """, format_traceback())



from cone.app.browser import render_main_template
from cone.app.browser.login import login_view
from cone.app.browser.utils import format_traceback
from cone.tile import Tile
from cone.tile import tile
from pyramid.httpexceptions import HTTPForbidden
from pyramid.httpexceptions import HTTPNotFound
from pyramid.response import Response
from pyramid.view import view_config
import json


###############################################################################
# Internal Server Error
###############################################################################

ERROR_PAGE = """
<h1>An error occured</h1>
<p>
  <a href="/">HOME</a>
  <hr />
</p>
%(error)s
"""


@view_config(context=Exception)
def internal_server_error(request):
    """Internal server error view.
    """
    tb = format_traceback()
    if not request.is_xhr:
        # XXX: request.response.status = 500
        #      return {'errors': error_dict}
        return Response(ERROR_PAGE % {'error': tb})
    # XXX: Check request relates to bdajax action
    #      Check if json request and modify response as needed
    from cone.app.browser.ajax import (
        AjaxContinue,
        AjaxMessage,
    )
    continuation = AjaxContinue([AjaxMessage(tb, 'error', None)]).definitions
    ret = {
        'mode': 'NONE',
        'selector': 'NONE',
        'payload': '',
        'continuation': continuation,
    }
    response = Response(json.dumps(ret))
    response.content_type = 'application/json'
    return response


###############################################################################
# Unauthorized
###############################################################################

@tile(name='unauthorized', path='templates/unauthorized.pt', permission='login')
class UnauthorizedTile(Tile):
    """Unauthorized tile.
    """


@view_config(context=HTTPForbidden, accept='text/html')
def forbidden_view(request):
    """Unauthorized view.
    """
    model = request.context
    if not request.authenticated_userid:
        return login_view(model, request)
    return render_main_template(model, request, contenttile='unauthorized')


@view_config(
    context=HTTPForbidden,
    accept='application/json',
    renderer='json')
def json_forbidden_view(request):
    request.response.status = 403
    return {}


###############################################################################
# Not Found
###############################################################################

@tile(name='not_found', path='templates/not_found.pt', permission='login')
class NotFoundTile(Tile):
    """Not Found tile.
    """


@view_config(context=HTTPNotFound, accept='text/html')
def not_found_view(request):
    """Not Found view.
    """
    model = request.context
    return render_main_template(model, request, contenttile='not_found')


@view_config(
    context=HTTPNotFound,
    accept='application/json',
    renderer='json')
def json_not_found_view(request):
    request.response.status = 404
    return {}



from cone.app.browser import render_main_template
from cone.app.browser.actions import LinkAction
from cone.app.browser.contextmenu import context_menu
from cone.app.browser.utils import make_url
from cone.tile import register_tile
from cone.tile import tile
from zope.interface import Interface
import sys


class content_view_tile(tile):
    """Extended tile decorator for registering content tiles.

    Additionally registers a view with the same name as the tile traversable
    via browser URL.
    """

    def create_content_view(self, ob):
        def content_view(model, request):
            return render_main_template(model, request, contenttile=self.name)

        view_name = '{}_content_view'.format(self.name)
        content_view.__doc__ = (
            'Dynamically created by '
            'cone.app.browser.content.content_view_tile'
        )
        content_view.__name__ = view_name
        content_view.__qualname__ = view_name
        content_view.__module__ = ob.__module__
        module = sys.modules[ob.__module__]
        setattr(module, view_name, content_view)

        def callback(context, name, ob_):
            config = context.config.with_package(info.module)
            config.add_view(
                view=ob_,
                name=self_.name,
                context=self_.interface,
                permission=self_.permission
            )
        self_ = self
        info = self.venusian.attach(
            content_view,
            callback,
            category='pyramid',
            depth=2
        )

    def __call__(self, ob):
        self.create_content_view(ob)

        def callback(context, name, ob_):
            register_tile(
                name=self_.name,
                path=self_.path,
                attribute=self_.attribute,
                interface=self_.interface,
                class_=ob_,
                permission=self_.permission,
                strict=self_.strict
            )
        self_ = self
        self.venusian.attach(ob, callback, category='pyramid', depth=1)
        return ob


class ContentViewAction(LinkAction):
    """View action for the contentviews group of the contextmenu.

    Gets created by ``content_view_action`` decorator.
    """

    def __init__(self, name, interface=None,
                 permission=None, text=None, icon=None):
        self.name = name
        self.interface = interface
        self.permission = permission
        self.text = text
        self.icon = icon

    @property
    def display(self):
        permitted = True
        if self.permission:
            permitted = self.permitted(self.permission)
        if self.interface:
            if issubclass(self.interface, Interface):
                return self.interface.providedBy(self.model) and permitted
            return isinstance(self.model, self.interface) and permitted
        return permitted

    @property
    def selected(self):
        return self.action_scope == self.name

    @property
    def href(self):
        return make_url(self.request, node=self.model, resource=self.name)


class content_view_action(object):
    """Decorator providing a shortcut for creating a view action and
    registering in the contentviews group of the contextmenu.
    """

    def __init__(self, name, tilename=None, interface=None,
                 permission=None, text=None, icon=None):
        self.name = name
        self.tilename = tilename if tilename is not None else name
        self.interface = interface
        self.permission = permission
        self.text = text
        self.icon = icon

    def __call__(self, ob):
        context_menu['contentviews'][self.name] = ContentViewAction(
            name=self.tilename,
            interface=self.interface,
            permission=self.permission,
            text=self.text,
            icon=self.icon
        )
        return ob



from cone.app import cfg
from cone.app import layout_config
from cone.app.browser.actions import get_action_context
from cone.app.browser.actions import LinkAction
from cone.app.browser.ajax import ajax_continue
from cone.app.browser.ajax import AjaxEvent
from cone.app.browser.utils import format_date
from cone.app.browser.utils import make_query
from cone.app.browser.utils import make_url
from cone.app.browser.utils import node_icon
from cone.app.interfaces import IApplicationNode
from cone.app.interfaces import ILayout
from cone.app.interfaces import INavigationLeaf
from cone.app.interfaces import IWorkflowState
from cone.app.model import AppRoot
from cone.app.ugm import principal_data
from cone.app.ugm import ugm_backend
from cone.app.utils import node_path
from cone.tile import render_template
from cone.tile import render_tile
from cone.tile import Tile
from cone.tile import tile
from node.utils import LocationIterator
from node.utils import safe_decode
from odict import odict
from pyramid.i18n import get_localizer
from pyramid.i18n import TranslationStringFactory
import warnings


_ = TranslationStringFactory('cone.app')


@tile(name='logo', path='templates/logo.pt', permission='login')
class LogoTile(Tile):
    """Tile rendering the logo.
    """


@tile(name='livesearch', path='templates/livesearch.pt', permission='login')
class LivesearchTile(Tile):
    """Tile rendering the live search.
    """


@tile(name='footer', path='templates/footer.pt', permission='login')
class FooterTile(Tile):
    """Tile rendering the page footer.
    """


@tile(name='insufficient_privileges',
      path='templates/insufficient_privileges.pt',
      permission='login')
class InsufficientPrivilegesTile(Tile):
    """Tile rendering insufficient privileges message.
    """


class ProtectedContentTile(Tile):
    """A tile rendering the loginform instead default if user is not
    authenticated.

    Supposed to be used for 'content' tiles if page should render login form
    instead of throwing Unauthorized. Needs to be registered for permission
    'login'.

    The permission of the tile itself can be defined via ``content_permission``
    """
    content_permission = 'view'

    def __call__(self, model, request):
        if not request.authenticated_userid:
            return render_tile(model, request, 'loginform')
        if not request.has_permission(self.content_permission, model):
            return render_tile(model, request, 'insufficient_privileges')
        return Tile.__call__(self, model, request)


class LayoutConfigTile(Tile):

    @property
    def config(self):
        model = self.model
        props = model.properties
        if props.default_child:
            model = model[props.default_child]
        if hasattr(model, 'layout'):
            warnings.warn(
                '``AppNode.layout`` is deprecated, use ``layout_config``',
                DeprecationWarning
            )
            return model.layout
        layout = self.request.registry.queryAdapter(model, ILayout, default=None)
        if layout:
            warnings.warn(
                '``ILayout`` adapter is deprecated, use ``layout_config``',
                DeprecationWarning
            )
            return layout
        return layout_config.lookup(model=model, request=self.request)


@tile(name='layout', path='templates/layout.pt', permission='login')
class Layout(LayoutConfigTile):
    """Main layout tile.
    """

    @property
    def contenttile(self):
        return get_action_context(self.request).scope


# personal tools action registry
personal_tools = odict()


class personal_tools_action(object):
    """Decorator defining a personaltools dropdown item.
    """

    def __init__(self, name):
        self.name = name

    def __call__(self, factory):
        personal_tools[self.name] = factory()
        return factory


@personal_tools_action(name='settings')
class ViewSettingsAction(LinkAction):
    text = _('settings', default='Settings')
    icon = 'ion-ios7-gear'
    event = 'contextchanged:#layout'
    path = 'href'

    @property
    def settings(self):
        root = self.model.root
        return root and root.get('settings') or None

    @property
    def target(self):
        return make_url(self.request, node=self.settings)

    href = target

    @property
    def display(self):
        settings = self.settings
        if not settings:
            return False
        if not self.request.has_permission('view', settings):
            return False
        if not len(settings):
            return False
        return True


@personal_tools_action(name='logout')
class LogoutAction(LinkAction):
    text = _('logout', default='Logout')
    icon = 'ion-log-out'
    action = 'logout:NONE:NONE'
    path_action = ''
    path_event = ''

    @property
    def href(self):
        return make_url(self.request, resource='logout')

    @property
    def path(self):
        return '/'


@tile(name='personaltools',
      path='templates/personaltools.pt',
      permission='view',
      strict=False)
class PersonalTools(Tile):
    """Personal tool tile.
    """

    @property
    def user(self):
        userid = self.request.authenticated_userid
        data = principal_data(userid)
        display_name = data.get(ugm_backend.user_display_attr)
        return display_name if display_name else userid

    @property
    def items(self):
        return [_(self.model, self.request) for _ in personal_tools.values()]


@tile(name='mainmenu',
      path='templates/mainmenu.pt',
      permission='view',
      strict=False)
class MainMenu(LayoutConfigTile):
    """Main Menu tile.

    * set ``skip_mainmenu`` on ``model.properties`` to ``True`` if node should
      not be displayed in mainmenu.

    * set ``mainmenu_display_children`` on ``model.properties`` to ``True`` if
      child nodes should be rendered as dropdown menu.

    * set ``mainmenu_empty_title`` on ``model.root.properties`` to ``True``
      if you want to render empty links in mainmenu.

    * If ``default_child`` is set on ``model.root.properties``, it is marked
      selected if no other current path is found.

    * If ``default_content_tile`` is set on ``model.root.properties``, it is
      considered in target link creation.
    """

    @property
    def menuitems(self):
        ret = list()
        path = node_path(self.model)
        if path:
            curpath = path[0]
        else:
            curpath = ''
        # work with ``self.model.root.keys()``, ``values()`` propably not works
        # due to the use of factory node.
        root = self.model.root
        root_props = root.properties
        # check for default child id if no curpath
        if not curpath and root_props.default_child:
            curpath = root_props.default_child
        # check whether to render mainmenu item title
        empty_title = root_props.mainmenu_empty_title
        # XXX: icons
        for key in root.keys():
            child = root[key]
            if self.ignore_node(child):
                continue
            selected = curpath == key
            item = self.create_item(child, empty_title, selected)
            if child.properties.mainmenu_display_children:
                item['children'] = self.create_children(child, selected)
            else:
                item['children'] = None
            ret.append(item)
        return ret

    def create_children(self, node, selected):
        children = list()
        path = node_path(self.model)
        if path and len(path) > 1 and path[0] == node.name:
            curpath = path[1]
        else:
            curpath = ''
        for key in node.keys():
            child = node[key]
            if self.ignore_node(child):
                continue
            selected = curpath == key
            item = self.create_item(child, False, selected)
            children.append(item)
        return children

    def ignore_node(self, node):
        if not IApplicationNode.providedBy(node):
            return True
        if node.properties.skip_mainmenu:
            return True
        if not self.request.has_permission('view', node):
            return True
        return False

    def create_item(self, node, empty_title, selected):
        md = node.metadata
        item = dict()
        item['id'] = node.name
        if empty_title:
            item['title'] = ''
            item['description'] = md.title
        else:
            item['title'] = md.title
            item['description'] = md.description
        item['url'] = make_url(self.request, node=node)
        query = make_query(contenttile=node.properties.default_content_tile)
        item['target'] = make_url(self.request, node=node, query=query)
        item['selected'] = selected
        item['icon'] = node_icon(node)
        return item


@tile(name='pathbar',
      path='templates/pathbar.pt',
      permission='view',
      strict=False)
class PathBar(Tile):

    @property
    def items(self):
        return self.items_for(self.model)

    def item_url(self, node):
        return make_url(self.request, node=node)

    def item_target(self, node):
        query = make_query(contenttile=node.properties.default_content_tile)
        return make_url(self.request, node=node, query=query)

    def items_for(self, model, breakpoint=None):
        items = list()
        for node in LocationIterator(model):
            title = node.metadata.title
            title = safe_decode(title) if title else title
            items.append({
                'title': title,
                'url': self.item_url(node),
                'target': self.item_target(node),
                'selected': False,
                'id': node.name,
                'default_child': node.properties.default_child,
            })
            if node is breakpoint:
                break
        items.reverse()
        ret = list()
        count = len(items)
        for i in range(count):
            default_child = items[i]['default_child']
            if default_child \
                    and i < count - 1 \
                    and default_child == items[i + 1]['id']:
                continue
            ret.append(items[i])

        # XXX: this is crap!
        if not breakpoint:
            ret[0]['title'] = 'Home'

        ret[-1]['selected'] = True
        return ret


@tile(name='navtree',
      path='templates/navtree.pt',
      permission='view',
      strict=False)
class NavTree(Tile):
    """Navigation tree tile.
    """

    @property
    def title(self):
        navroot = self.navroot
        default = _('navigation', default='Navigation')
        if self.model.root is navroot:
            return default
        return navroot.metadata.get('title', default)

    @property
    def navroot(self):
        model = self.model
        root = model.root
        while model is not root:
            if model.properties.is_navroot:
                return model
            model = model.parent
        return root

    def navtreeitem(self, title, url, target, path, icon, css=''):
        item = dict()
        item['title'] = title
        item['url'] = url
        item['target'] = target
        item['selected'] = False
        item['path'] = path
        item['icon'] = icon
        item['css'] = css
        item['showchildren'] = False
        item['children'] = list()
        return item

    def fillchildren(self, model, path, tree):
        if INavigationLeaf.providedBy(model):
            return
        curpath = None
        if path:
            curpath = path[0]
        default_child = None
        if model.properties.default_child:
            if not curpath or curpath == model.properties.default_child:
                default_child = model[model.properties.default_child]
        if default_child and default_child.properties.hide_if_default:
            model = default_child
            default_child = None
            if path:
                path = path[1:]
                if path:
                    curpath = path[0]
        if default_child:
            if not curpath:
                curpath = model.properties.default_child
        for key in model:
            node = model[key]
            if not IApplicationNode.providedBy(node):
                continue
            if not self.request.has_permission('view', node):
                continue
            if not node.properties.in_navtree:
                continue
            title = node.metadata.title
            if title:
                title = safe_decode(title)
            url = make_url(self.request, node=node)
            query = make_query(contenttile=node.properties.default_content_tile)
            target = make_url(self.request, node=node, query=query)
            curnode = curpath == safe_decode(key)
            icon = node_icon(node)
            css = ''
            if IWorkflowState.providedBy(node):
                css = 'state-%s' % node.state
            child = self.navtreeitem(
                title, url, target, node_path(node), icon, css)
            child['showchildren'] = curnode
            if curnode:
                child['selected'] = True
                if default_child:
                    self.fillchildren(default_child, path[1:], child)
                else:
                    self.fillchildren(node, path[1:], child)
            else:
                selected_path = node_path(self.model)
                if default_child:
                    selected_path.append(default_child.name)
                selected = False
                # XXX: probably superfluous. keep as of cone.app 1.1
                # if selected_path == node_path(node):
                #     selected = True
                child['selected'] = selected
            tree['children'].append(child)

    def navtree(self):
        root = self.navtreeitem(None, None, None, '', None)
        model = self.navroot
        # XXX: default child
        path = node_path(self.model)[len(node_path(model)):]
        self.fillchildren(model, path, root)
        return root

    def rendertree(self, children, level=1):
        return render_template(
            'cone.app.browser:templates/navtree_recue.pt',
            model=self.model,
            request=self.request,
            context=self,
            children=children,
            level=level)


@tile(name='byline',
      path='templates/byline.pt',
      permission='view',
      strict=False)
class Byline(Tile):
    """Byline tile.
    """

    def format_date(self, dt):
        return format_date(dt)


@tile(name='content', interface=AppRoot, permission='login')
class RootContent(ProtectedContentTile):

    def render(self):
        default_child = self.model.properties.default_child
        if default_child:
            model = self.model[default_child]
            return render_tile(model, self.request, 'content')
        default_content_tile = self.model.properties.default_content_tile
        if default_content_tile:
            return render_tile(self.model, self.request, default_content_tile)
        return render_template(
            'cone.app.browser:templates/default_root.pt',
            model=self.model,
            request=self.request,
            context=self)


class LanguageTile(Tile):
    param_blacklist = [
        '_', '_LOCALE_', 'bdajax.action', 'bdajax.mode', 'bdajax.selector'
    ]

    def make_query(self, lang=None):
        params = dict()
        for k, v in self.request.params.items():
            if k in self.param_blacklist:
                continue
            params[k] = v
        params['lang'] = lang
        return make_query(**params)


language_names = {
    'en': _('lang_en', default='English'),
    'de': _('lang_de', default='German'),
    'fr': _('lang_fr', default='French'),
    'it': _('lang_it', default='Italian')
}


@tile(name='language',
      path='templates/language.pt',
      permission='login',
      strict=False)
class Language(LanguageTile):

    @property
    def show(self):
        return bool(cfg.available_languages)

    @property
    def languages(self):
        languages = list()
        localizer = get_localizer(self.request)
        for lang in cfg.available_languages:
            target = make_url(
                self.request,
                node=self.model,
                query=self.make_query(lang=lang)
            )
            title = localizer.translate(language_names.get(lang, lang.upper()))
            languages.append({
                'target': target,
                'icon': 'icon-lang-{}'.format(lang),
                'title': title
            })
        return languages


@tile(name='change_language', permission='login')
class ChangeLanguage(LanguageTile):

    @property
    def continuation(self):
        url = make_url(self.request, node=self.model, query=self.make_query())
        return [AjaxEvent(url, 'contextchanged', '#layout')]

    def render(self):
        lang = self.request.params['lang']
        response = self.request.response
        response.set_cookie('_LOCALE_', value=lang, max_age=31536000)
        ajax_continue(self.request, self.continuation)
        return u''
